"""
A simple repeat with two alternative endings (volta brackets).
"""

from pathlib import Path

from musicscore import Score, Chord
from unittest import TestCase


class TestLily45b(TestCase):
    def test_lily_45b_RepeatWithAlternatives(self):
        score = Score()
        part = score.add_part("p1")
        [part.add_chord(Chord(72, 4)) for _ in range(4)]
        part.get_measure(2).set_repeat_barline()
        part.get_measure(2).set_repeat_ending(number=1, type="start")
        part.get_measure(2).set_repeat_ending(number=1, type="stop")
        part.get_measure(3).set_repeat_ending(number=2, type="start")
        part.get_measure(3).set_repeat_ending(number=2, type="discontinue")
        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""Lyrics assigned to the voices of a piano staff containing two simple staves. Each staff is assigned exactly one
lyrics line."""

from pathlib import Path

from musicscore import Score, A, Chord, F, Lyrics
from unittest import TestCase


class TestLily61c(TestCase):
    def test_lily_61c_Lyrics_Pianostaff(self):
        score = Score()
        part = score.add_part("p1")
        chords_1 = [Chord(A(4), 1) for _ in range(4)]
        chords_2 = [Chord(F(3), 1) for _ in range(4)]
        Lyrics([("tra", "la", "li"), "ja!"]).add_to_chords(chords_1)
        Lyrics([("TRA", "LA", "LI"), "JA!"]).add_to_chords(chords_2)

        [part.add_chord(ch, staff_number=1) for ch in chords_1]
        [part.add_chord(ch, staff_number=2) for ch in chords_2]

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""
Staff changes in a piano staff. The voice from the second staff has some notes/chords on the first staff. The
final two chords have some notes on the first, some on the second staff.
"""

from pathlib import Path

from musicscore import Score, Chord, C, A, E, Time, G
from unittest import TestCase


class TestLily43a(TestCase):
    def test_lily_43a_MultiStaff_StaffChange(self):
        score = Score()
        part = score.add_part("p1")
        t = Time(4, 4)
        t.actual_signatures = [1, 2, 1, 2]
        part.add_measure(time=t)

        part.add_chord(Chord(0, 4), staff_number=1, voice_number=1)

        for index, m in enumerate([A(3), E(4), A(3), E(4), C(5), E(4), A(3), A(4)]):
            if index in [1, 3, 4, 5]:
                m.set_staff_number(1)
            part.add_chord(Chord(m, 0.5), staff_number=2, voice_number=1)

        part.add_chord(Chord(0, 4), staff_number=1, voice_number=1)
        for index, m in enumerate(
            [
                [C(3), E(3), G(3), C(4)],
                [C(4), E(4), G(4)],
                [C(3), E(3), G(3), C(4)],
                [G(3), C(4), E(4), G(4)],
            ]
        ):
            if index in [1]:
                for x in m:
                    x.set_staff_number(1)
            part.add_chord(Chord(m, 0.5), staff_number=2, voice_number=1)

        part.add_chord(Chord(0, 2), staff_number=2)
        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""
A simple piano staff
"""

from pathlib import Path

from musicscore import Score, F, Chord, B
from unittest import TestCase


class TestLily43a(TestCase):
    def test_lily_43a_PianoStaff(self):
        score = Score()
        part = score.add_part("p1")
        part.add_chord(Chord(F(4), 4), staff_number=1)
        part.add_chord(Chord(B(2), 4), staff_number=2)

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



import cProfile
from pathlib import Path

import random

from musicscore import Score, Part, QuarterDuration, Chord


def p():
    """
    Tester creates a timewise score
    """
    s = Score()
    """
    He adds a list of complicated quarter_durations
    """
    p = s.add_child(Part("p1"))

    random.seed(11)
    quarter_durations = []
    while sum(quarter_durations) < 40:
        quarter_durations.append(
            QuarterDuration(random.random() + random.randint(0, 3))
        )
    quarter_durations.append(44 - sum(quarter_durations))
    """
    He sets the possible subdivisions for all beats with quarter_duration and omits quintuplets and septuplets:
    """
    s.set_possible_subdivisions([2, 3, 4, 6, 8])
    """
    He adds chords with quarter_durations and writes each quarter_duration as lyrics
    """
    for qd in quarter_durations:
        ch = Chord(midis=71, quarter_duration=qd)
        ch.add_lyric(round(float(ch.quarter_duration), 2))
        p.add_chord(ch)
    """
    He sets part's get_quantized attribute to True.
    """
    p.quantize = True
    """
    # ... and exports the xml
    # """
    xml_path = Path(__file__).with_suffix(".xml")
    s.export_xml(xml_path)


cProfile.run("p()", sort="cumtime")



"""
Four multi-measure rests: 3 measures, 15 measures, 1 measure, and 12 measures.
"""

from pathlib import Path

from musicscore import Score
from unittest import TestCase


class TestLily61a(TestCase):
    def test_lily_61a_Lyrics(self):
        score = Score()
        score.add_part("p1")
        score.set_multi_measure_rest(1, 3)
        score.set_multi_measure_rest(4, 19)
        score.set_multi_measure_rest(21, 32)

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from fractions import Fraction
from pathlib import Path
import random
from unittest import TestCase

from musicscore.chord import Chord
from musicscore.part import Part
from musicscore.quarterduration import QuarterDuration
from musicscore.score import Score
from musicscore.tests.util import XMLTestCase
from musicscore.util import lcm


class TestQuantization(TestCase):
    def test_get_and_set_possible_subdivisions(self):
        s = Score()
        p = s.add_child(Part("p1"))
        m = p.add_measure()
        s = m.get_children()[0]
        v = s.get_children()[0]
        beats = v.get_children()
        for beat in beats:
            assert beat.get_possible_subdivisions() == [2, 3, 4, 5, 6, 7, 8]
        beats[0].quarter_duration = 1 / 2
        assert beats[0].get_possible_subdivisions() == [2, 3, 4, 5]
        beats[0].quarter_duration = 1 / 4
        assert beats[0].get_possible_subdivisions() == [2, 3]
        beats[0].quarter_duration = 1 / 8
        assert beats[0].get_possible_subdivisions() is None
        v.set_possible_subdivisions([2, 3, 4, 6], beat_quarter_duration=1)
        beats[1].set_possible_subdivisions([2, 4])
        assert beats[0].get_possible_subdivisions() is None
        assert beats[1].get_possible_subdivisions() == [2, 4]
        for beat in beats[2:]:
            assert beat.get_possible_subdivisions() == [2, 3, 4, 6]

        with self.assertRaises(ValueError):
            beats[1].set_possible_subdivisions([2, 4], beat_quarter_duration=0.5)

    def test_quantize_attribute(self):
        s = Score()
        p = s.add_child(Part("p1"))
        m = p.add_measure()

        st = m.get_children()[0]

        v = st.get_children()[0]
        beats = v.get_children()

        assert s.get_quantized is False
        assert p.get_quantized is False
        assert m.get_quantized is False
        assert st.get_quantized is False
        assert v.get_quantized is False
        for b in beats:
            assert b.get_quantized is False

        s.get_quantized = True
        assert p.get_quantized is True
        assert m.get_quantized is True
        assert st.get_quantized is True
        assert v.get_quantized is True
        for b in beats:
            assert b.get_quantized is True

        st.get_quantized = False
        assert s.get_quantized is True
        assert p.get_quantized is True
        assert m.get_quantized is True
        assert st.get_quantized is False
        assert v.get_quantized is False
        for b in beats:
            assert b.get_quantized is False

        p.get_quantized = False
        m.get_quantized = True
        st.get_quantized = False
        v.get_quantized = True
        beats[0].get_quantized = False
        beats[1].get_quantized = True
        beats[2].get_quantized = False
        beats[3].get_quantized = True

        assert s.get_quantized is True
        assert p.get_quantized is False
        assert m.get_quantized is True
        assert st.get_quantized is False
        assert v.get_quantized is True

    def test_simple_quantization(self):
        s = Score()
        p = s.add_child(Part("p1"))
        p.add_measure()
        p.add_chord(Chord(60, 0.2))
        p.add_chord(Chord(60, 3.8))
        assert [ch.quarter_duration for ch in p.get_chords()] == [0.2, 0.8, 3]
        b = p.get_measure(1).get_staff(1).get_voice(1).get_beat(1)
        b.set_possible_subdivisions(subdivisions=[2, 3, 4])
        b.quantize_quarter_durations()
        assert [
            ch.quarter_duration
            for ch in p.get_measure(1).get_staff(1).get_voice(1).get_chords()
        ] == [0.25, 0.75, 3]
        assert [
            ch.offset for ch in p.get_measure(1).get_staff(1).get_voice(1).get_chords()
        ] == [0, 0.25, 0]

    def test_complex_quantization(self):
        s = Score()
        s.set_possible_subdivisions([2, 3, 4, 6, 8])
        p = s.add_child(Part("p1"))
        random.seed(11)
        quarter_durations = []
        while sum(quarter_durations) < 40:
            quarter_durations.append(
                QuarterDuration(random.random() + random.randint(0, 3))
            )
        quarter_durations.append(44 - sum(quarter_durations))

        for qd in quarter_durations:
            p.add_chord(Chord(midis=60, quarter_duration=qd))
        for beat in [
            b
            for m in p.get_children()
            for st in m.get_children()
            for v in st.get_children()
            for b in v.get_children()
        ]:
            beat.quantize_quarter_durations()
        # p.get_quantized = True
        for measure in p.get_children():
            v = measure.get_staff(1).get_voice(1)
            for b in v.get_children():
                qds = [ch.quarter_duration for ch in b.get_children()]
                div = lcm([qd.denominator for qd in qds])
                if div != 1:
                    assert div in [2, 3, 4, 6, 8]

    def test_add_short_chord(self):
        s = Score()
        p = s.add_part("part-1")
        p.add_chord(Chord(60, 9 / 20))
        p.get_quantized = True
        p.finalize()
        self.assertEqual(
            [chord.quarter_duration for chord in p.get_chords()], [3 / 7, 4 / 7, 3 / 1]
        )

    def test_part_quantization(self):
        s = Score()
        s.set_possible_subdivisions([2, 3, 4, 6, 8])
        p = s.add_child(Part("p1"))
        for qd in [2, 2 / 7, 1, 5 / 7]:
            p.add_chord(Chord(60, qd))
        p.get_quantized = True
        p.finalize()
        assert [ch.quarter_duration for ch in p.get_chords()] == [
            2,
            1 / 4,
            3 / 4,
            1 / 4,
            3 / 4,
        ]
        assert p.get_measure(1).get_divisions() == 4
        assert [ch.notes[0].xml_duration.value_ for ch in p.get_chords()] == [
            8,
            1,
            3,
            1,
            3,
        ]


path = Path(__file__)


class TestQuantizationWithAfterGraceNotes(XMLTestCase):
    def test_quantization_with_grace_notes(self):
        s = Score()
        s.get_quantized = True
        p = s.add_child(Part("p1"))
        p.set_possible_subdivisions([2, 3, 4])
        p.add_measure(time=(1, 4))
        for qd in [0, 0, 1 / 2, 0, 0, 1 / 2 - 1 / 8, 1 / 16, 1 / 16]:
            p.add_chord(Chord(60, qd))
        b = p.get_beat(1, 1, 1, 1)
        b.quantize_quarter_durations()
        self.assertEqual(
            [ch.quarter_duration for ch in b.get_chords()],
            [0, 0, 1 / 2, 0, 0, 1 / 2, 0, 0],
        )
        s.finalize()
        with self.file_path(path, "with_grace_notes") as xml_path:
            s.export_xml(xml_path)

    def test_quantization_with_grace_notes_complex(self):
        s = Score()
        s.get_quantized = True
        p = s.add_child(Part("p1"))
        durations = [
            Fraction(128, 675),
            Fraction(64, 135),
            Fraction(64, 675),
            Fraction(64, 225),
            Fraction(256, 675),
            Fraction(256, 675),
            Fraction(1024, 3375),
            Fraction(512, 3375),
            Fraction(256, 3375),
            Fraction(256, 1125),
            Fraction(128, 3375),
            Fraction(256, 3375),
            Fraction(128, 1125),
            Fraction(512, 3375),
            Fraction(128, 675),
            Fraction(256, 3375),
            Fraction(64, 1125),
            Fraction(64, 675),
            Fraction(128, 3375),
            Fraction(64, 3375),
            Fraction(64, 375),
            Fraction(64, 1125),
            Fraction(256, 1125),
            Fraction(64, 225),
            Fraction(128, 1125),
            Fraction(16, 375),
        ]
        for qd in durations:
            p.add_chord(Chord(60, qd))
        # s.finalize()
        with self.file_path(path, "with_grace_notes_complex") as xml_path:
            s.export_xml(xml_path)



from typing import List, Union, Optional
from fractions import Fraction
import numbers

__all__ = ["QuarterDuration", "QuarterDurationMixin"]

from musicscore.config import (
    NOTETYPES,
    BEATWISE_EXCEPTIONS,
    DOTEDTUPLETRATIO,
    TYPEANDDOTEXCEPTIONS,
)
from musicscore.exceptions import QuarterDurationIsNotWritable


class QuarterDuration(numbers.Rational):
    """
    A Class specifically designed for durations measured in quarters. The core of this class is a value of type ''fractions.Fraction'' with a
    denominator limit of 1000, thus it can manage conversion of floats to fractions without usual inaccuracies of quintuples etc. See
    value property for more information.
    QuarterDuration has all needed magic methods for numeral comparison and conversion.
    """

    def __init__(self, *value):
        self._value = None
        self.value = value
        self._beat_subdivision = None
        self._beat_quarter_duration = 1
        self._type_and_dots = None

    def _get_beatwise_sections(
        self,
        beats: List["Beat"],
        offset: Union[int, float, "QuarterDuration", "Fraction"] = 0,
    ):
        """
        :param beats:
        :param offset: offset in the first beat
        :return: [sections as list of QuarterDurations, leftover as QuarterDruation] leftover is the remaining quarter_duration which
                 exceeds the sum of all beats quarter durations. If there is no left over the second value in the list is None.
                 offset=0.5, beats=[Beat(1), Beat(1), Beat(1), Beat(1)] => [[0.5, 3], None]
                 See tests for more examples. In BEATWISE_EXCEPTIONS exceptions can be declared.
        """

        def _check_for_exception():
            exception = BEATWISE_EXCEPTIONS.get(offset)
            if exception:
                out = exception.get(sum(output[0]))
                if out:
                    output[0] = [QuarterDuration(value) for value in out]
                return output

        output = [None, [], None]
        if offset:
            output[0] = beats[0].quarter_duration - offset
            beats.pop(0)
        leftover = self - output[0] if output[0] is not None else self
        out_of_reach = leftover - sum(b.quarter_duration for b in beats)
        if out_of_reach > 0:
            leftover -= out_of_reach
            output[2] = out_of_reach
        for beat in beats:
            if leftover >= beat.quarter_duration:
                current_value = beat.quarter_duration
                leftover -= current_value
            else:
                current_value = leftover
                leftover = 0
            if not output[1]:
                output[1] = [current_value]
            else:
                if _is_writable(output[1][-1] + current_value):
                    output[1][-1] += current_value
                else:
                    output[1].append(current_value)
            if leftover == 0:
                break
        if output[0]:
            output = [[output[0], *output[1]], output[2]]
        else:
            output = [output[1], output[2]]

        exception = _check_for_exception()
        if exception:
            return exception

        return output

    def _get_type_and_dots(self):
        if self.value == 0:
            return None, 0
        if not self.beat_subdivision:
            self.beat_subdivision = self.denominator
        try:
            type_and_dots = (
                TYPEANDDOTEXCEPTIONS.get(self.beat_quarter_duration)
                .get(self.beat_subdivision)
                .get(self.as_integer_ratio())
            )
            if type_and_dots:
                return type_and_dots
        except AttributeError:
            pass
        type = NOTETYPES.get(self.as_integer_ratio())
        if type:
            return type, 0
        else:
            qd = QuarterDuration(self.value * 2 / 3)
            type = NOTETYPES.get(qd.as_integer_ratio())
            if type:
                return type, 1
            else:
                qd = QuarterDuration(self.value * 4 / 7)
                type = NOTETYPES.get(qd.as_integer_ratio())
                if type:
                    return type, 2
                else:
                    raise QuarterDurationIsNotWritable(
                        f"quarter duration {self} is not writable."
                    )

    @property
    def beat_subdivision(self):
        return self._beat_subdivision

    @beat_subdivision.setter
    def beat_subdivision(self, val):
        self._beat_subdivision = val
        self._type_and_dots = None

    @property
    def beat_quarter_duration(self):
        return self._beat_quarter_duration

    @beat_quarter_duration.setter
    def beat_quarter_duration(self, val):
        self._beat_quarter_duration = val
        self._type_and_dots = None

    @property
    def denominator(self):
        """
        :return: Fraction's denominator.
        :rtype: int

        >>> QuarterDuration(1, 6).denominator
        6
        """
        return self.value.denominator

    @property
    def numerator(self):
        """
        :return: Fraction's numerator.
        :rtype: int

        >>> QuarterDuration(1, 6).numerator
        1
        """
        return self.value.numerator

    @property
    def type_and_dots(self):
        if self._type_and_dots is None:
            self._type_and_dots = self._get_type_and_dots()
        return self._type_and_dots

    @property
    def value(self):
        """
        :return: QuarterDuration's value
        :rtype: fractions.Fraction with limit_denominator(1000)

        >>> QuarterDuration(3, 7).value
        Fraction(3, 7)
        >>> QuarterDuration(0.2).value
        Fraction(1, 5)
        >>> QuarterDuration(Fraction(1, 5)).value
        Fraction(1, 5)
        >>> QuarterDuration(1/5).value
        Fraction(1, 5)
        """
        return self._value

    @value.setter
    def value(self, val):
        if val is None or val == ():
            self._value = None
        elif isinstance(val, Fraction):
            self._value = val
        elif isinstance(val, str):
            self._value = Fraction(val).limit_denominator(1000)
        elif hasattr(val, "__iter__"):
            if len(val) == 1:
                self._value = Fraction(val[0]).limit_denominator(1000)
            elif len(val) == 2:
                self._value = Fraction(*val).limit_denominator(1000)
            else:
                raise ValueError
        else:
            try:
                self._value = Fraction(val).limit_denominator(1000)
            except TypeError:
                raise TypeError("Wrong type for QuarterDuration.value")

    def as_integer_ratio(self):
        """
        :return: (numerator, denominator)
        :rtype: tuple

        >>> QuarterDuration(1, 5).as_integer_ratio()
        (1, 5)
        """
        return self.value.as_integer_ratio()

    def get_number_of_dots(self) -> Optional[int]:
        """
        :return: Number of note dots associated with quarter duration
        """
        return self.type_and_dots[1]

    def get_tuplet_ratio(self) -> Optional[tuple]:
        if self.value == 0:
            return None
        if not self.beat_subdivision:
            self.beat_subdivision = self.denominator
        if self.beat_quarter_duration % 3 == 0:
            if self.beat_subdivision > 9:
                raise NotImplementedError(
                    "Beats with dotted quarter duration and subdivision > 9"
                )
            else:
                tupletratio = DOTEDTUPLETRATIO.get(self.beat_subdivision)
                if tupletratio:
                    return self.beat_subdivision, tupletratio
                else:
                    return None
        else:
            if self.beat_subdivision < 3:
                return None
            elif self.beat_subdivision > 64:
                raise NotImplementedError("Beats subdivision > 64")
            normal_notes = [2, 4, 8, 16, 32]
            if self.beat_subdivision in normal_notes:
                return None
            else:
                for normal in reversed(normal_notes):
                    if self.beat_subdivision > normal:
                        return self.beat_subdivision, normal
        raise NotImplementedError(
            f"Quarter duration {self} in a beat with {self.beat_subdivision} and quarter duration {self.beat_quarter_duration}"
        )

    def get_type(self) -> Optional[str]:
        """
        :return: Note type associated with quarter duration
        """
        return self.type_and_dots[0]

    def __repr__(self):
        return f"{self.value.numerator}/{self.value.denominator}"

    def __str__(self):
        return f"QuarterDuration: {str(self.value)}"

    def __abs__(self):
        return QuarterDuration(self.value.__abs__())

    def __add__(self, other):
        return QuarterDuration(self.value.__add__(_convert_other(other)))

    def __ceil__(self):
        return QuarterDuration(self.value.__ceil__())

    def __eq__(self, other):
        return self.value.__eq__(_convert_other(other))

    def __floor__(self):
        return QuarterDuration(self.value.__floor__())

    def __floordiv__(self, other):
        return QuarterDuration(self.value.__floordiv__(_convert_other(other)))

    def __gt__(self, other):
        return self.value.__gt__(_convert_other(other))

    def __ge__(self, other):
        return self.value.__ge__(_convert_other(other))

    def __hash__(self):
        return self.value.__hash__()

    def __le__(self, other):
        return self.value.__le__(_convert_other(other))

    def __lt__(self, other):
        return QuarterDuration(self.value.__lt__(_convert_other(other)))

    def __mod__(self, other):
        return QuarterDuration(self.value.__mod__(_convert_other(other)))

    def __mul__(self, other):
        return QuarterDuration(self.value.__mul__(_convert_other(other)))

    def __neg__(self):
        return self.value.__neg__()

    def __pos__(self):
        return self.value.__pos__()

    def __pow__(self, power):
        return QuarterDuration(self.value.__pow__(power))

    def __radd__(self, other):
        return QuarterDuration(self.value.__radd__(_convert_other(other)))

    def __rfloordiv__(self, other):
        return QuarterDuration(self.value.__rfloordiv__(_convert_other(other)))

    def __rmod__(self, other):
        return QuarterDuration(self.value.__rmod__(_convert_other(other)))

    def __rmul__(self, other):
        return QuarterDuration(self.value.__rmul__(_convert_other(other)))

    def __round__(self, n=None):
        return QuarterDuration(self.value.__round__(n))

    def __rpow__(self, other):
        return QuarterDuration(self.value.__rpow__(_convert_other(other)))

    def __rtruediv__(self, other):
        return QuarterDuration(self.value.__rtruediv__(_convert_other(other)))

    def __truediv__(self, other):
        return QuarterDuration(self.value.__truediv__(_convert_other(other)))

    def __trunc__(self):
        return self.value.__trunc__()

    def __copy__(self):
        return self.__class__(self.value)

    def __deepcopy__(self, memodict={}):
        return self.__class__(self.value)


def _is_writable(quarter_duration: Union[float, int, Fraction, "QuarterDuration"]):
    """
    Function to check if a quarter duration is writable or must be split into two durations.

    :param quarter_duration:
    :return: boolean

    >>> _is_writable(5)
    False
    >>> _is_writable(7/8)
    False
    >>> _is_writable(3/8)
    True
    """
    writables = {
        1 / 64,
        1 / 32,
        3 / 64,
        1 / 16,
        3 / 32,
        1 / 8,
        3 / 16,
        1 / 4,
        3 / 8,
        1 / 2,
        3 / 4,
        1,
        3 / 2,
        2,
        3,
        4,
        6,
        8,
        12,
    }

    if quarter_duration in writables:
        return True
    else:
        return False


def _convert_other(other) -> bool:
    if isinstance(other, QuarterDuration):
        return other.value
    try:
        return Fraction(other).limit_denominator(1000)
    except TypeError:
        return False


class QuarterDurationMixin:
    """
    Mixin for all Classes with a quarter_duration. Used in :obj:`~musicscore.note.Note`, :obj:`~musicscore.chord.Chord` and
    :obj:`~musicscore.beat.Beat`
    """

    def __init__(self, quarter_duration=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._quarter_duration = None
        self.quarter_duration = quarter_duration

    def _set_quarter_duration(self, val):
        if isinstance(val, QuarterDuration):
            self._quarter_duration = val
        else:
            _check_quarter_duration_value(val)
            self._quarter_duration = QuarterDuration(val)

    @property
    def quarter_duration(self) -> QuarterDuration:
        """
        Set and get the duration measured in quarters.

        Setting value can be of types ``int``, ``float``, ``fractions.Fraction``, :obj:`~musicscore.quarterduration.QuarterDuration`
        """
        return self._quarter_duration

    @quarter_duration.setter
    def quarter_duration(self, val):
        if val is not None:
            self._set_quarter_duration(val)
        else:
            self._quarter_duration = None


def _check_quarter_duration_value(val):
    if (
        not isinstance(val, int)
        and not isinstance(val, float)
        and not isinstance(val, Fraction)
        and not isinstance(val, QuarterDuration)
    ):
        raise TypeError(f"Wrong type for quarter duration {val}: {type(val)}")

    if val < 0:
        raise ValueError()

    return True



"""The number of staff lines can be modified by using the staff-lines child of the staff-details attribute. This can
happen globally (the first staff has one line globally) or during the part at the beginning of a measure and even
inside a measure (the second part has 5 lines initially, 4 at the beginning of the second measure, and 3 starting in
the middle of the third measure)."""

from pathlib import Path

from musicscore import Score, Chord, G
from unittest import TestCase
from musicxml import XMLStaffDetails

"""
finale changes staff number of lines only ones. The next changes have no effect.
"""


class TestLily14a(TestCase):
    def test_lily_14a_StaffDetails_LineChanges(self):
        score = Score()
        p1 = score.add_part("p1")
        p2 = score.add_part("p2")
        p1.name = "Part 1"
        p2.name = "Part 2"

        [p1.add_chord(Chord(G(4), 4)) for _ in range(4)]
        [p2.add_chord(Chord(G(4), 2)) for _ in range(8)]

        sd = p1.get_measure(1).xml_attributes.xml_staff_details = XMLStaffDetails()
        sd.xml_staff_lines = 1

        sd = p2.get_measure(1).xml_attributes.xml_staff_details = XMLStaffDetails()
        sd.xml_staff_lines = 4

        sd = p2.get_measure(2).xml_attributes.xml_staff_details = XMLStaffDetails()
        sd.xml_staff_lines = 3

        sd = p2.get_measure(3).xml_attributes.xml_staff_details = XMLStaffDetails()
        sd.xml_staff_lines = 2
        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from pathlib import Path

from musicscore import Score, C, D, E, F, Chord, Midi
from unittest import TestCase

"""
Some microtones: c flat-and-a-half, d half-flat, e half-sharp, f sharp-and-a half. Once in the
lower and once in the upper region of the staff.
"""


class TestLily01d(TestCase):
    def test_lily_01d_Pitches_Microtones(self):
        score = Score("Microtones")
        p = score.add_part("microtones")
        midis = [Midi(60 - 3 / 2), Midi(60 + 3 / 2), Midi(64 + 1 / 2), Midi(65 + 3 / 2)]
        midis[0].accidental.mode = "force-flat"
        midis[1].accidental.mode = "force-flat"
        midis += [
            C(5, "three-quarters-flat"),
            D(5, "quarter-flat"),
            E(5, "quarter-sharp"),
            F(5, "three-quarters-sharp"),
        ]

        for midi in midis:
            p.add_chord(Chord(midi, 1))

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



import cProfile
import timeit
from contextlib import redirect_stdout
from pathlib import Path

from musicxml.xmlelement.xmlelement import XMLScorePartwise


def hello_xml_score():
    XMLScorePartwise()


def time_hello_xml_score():
    print(timeit.timeit(hello_xml_score, number=1000))


with open(str(Path(__file__)) + "xml_score.txt", "+w") as f:
    with redirect_stdout(f):
        cProfile.run("time_hello_xml_score()", sort="tottime")



from unittest import TestCase
from unittest.mock import patch

from musicscore.chord import Chord
from musicscore.measure import Measure
from musicscore.part import Part
from musicscore.staff import Staff
from musicscore.voice import Voice


class TestStaff(TestCase):
    def test_staff_init(self):
        st = Staff()
        assert st.number is None
        assert st.xml_object.value_ == 1
        st = Staff(number=3)
        assert st.xml_object.value_ == 3
        assert st.number == 3
        st.number = 2
        assert st.xml_object.value_ == 2
        st.xml_object.value_ = 1
        assert st.number == 1
        assert st.xml_object.value_ == 1

    @patch("musicscore.measure.Measure")
    def test_add_child(self, mock_measure):
        st = Staff()
        st._parent = mock_measure
        assert [child.number for child in st.get_children()] == []
        st.add_child(Voice())
        assert [child.number for child in st.get_children()] == [1]
        st.add_child(Voice())
        assert len(st.get_children()) == 2
        assert [child.number for child in st.get_children()] == [1, 2]

    @patch("musicscore.measure.Measure")
    def test_add_voice(self, mock_measure):
        st = Staff()
        st._parent = mock_measure
        v = st.add_voice()
        assert st.get_children()[-1] == v
        assert len(st.get_children()) == 1
        assert v.number == 1
        v = st.add_voice()
        assert st.get_children()[-1] == v
        assert len(st.get_children()) == 2
        assert v.number == 2

        v = st.add_voice(5)
        assert v.number == 5
        assert len(st.get_children()) == 5
        assert [v.number for v in st.get_children()] == [1, 2, 3, 4, 5]

    def test_get_previous_staff(self):
        p = Part("P1")
        m1 = p.add_measure()
        m2 = p.add_measure()
        m1.add_staff(2)
        m2.add_staff(2)
        st11, st12 = m1.get_children()
        st21, st22 = m2.get_children()
        assert st11.get_previous_staff() is None
        assert st12.get_previous_staff() is None
        assert st21.get_previous_staff() == st11
        assert st22.get_previous_staff() == st12

    def test_get_last_steps_with_accidentals(self):
        m1 = Measure(1)
        m1._add_chord(Chord(midis=[61, 62, 63], quarter_duration=2))
        m1._add_chord(Chord(midis=[63, 64, 66], quarter_duration=2))
        assert m1.get_staff(1).get_last_pitch_steps_with_accidentals() == {"E", "F"}

    def test_staff_clef(self):
        st = Staff()
        assert st.clef is None

    def test_get_staff_number_from_midi(self):
        part = Part("p1")
        ch = Chord([60, 70, 80], 2)
        part.add_chord(ch)
        ch.midis[0].set_staff_number(2)
        ch.midis[1].set_staff_number(2)
        part.finalize()

        for index, n in enumerate(ch.notes):
            if index in [0, 1]:
                assert n.xml_object.xml_staff.value_ == 2
            else:
                assert not n.xml_object.xml_staff



import copy
from unittest import skip, TestCase
from unittest.mock import Mock

from fractions import Fraction

from musicscore import BassClef, Score, Part
from musicscore.accidental import Accidental
from musicscore.beat import Beat
from musicscore.chord import Chord, _split_copy, _group_chords, GraceChord, Rest
from musicscore.exceptions import (
    ChordHasNoParentBeamError,
    DeepCopyException,
    ChordException,
    MusicTreeException,
    ChordAddXPlacementException,
    RestCannotSetMidiError,
    RestWithDisplayStepHasNoDisplayOctave,
    RestWithDisplayOctaveHasNoDisplayStep,
    GraceChordCannotHaveGraceNotesError,
    AlreadyFinalizedError,
    ChordAlreadyHasNotesError,
    ChordTypeNotSetError,
    ChordNumberOfDotsNotSetError,
    ChordTestError,
    ChordParentBeamError,
)
from musicscore.midi import Midi
from musicscore.quarterduration import QuarterDuration
from musicscore.tests.util import ChordTestCase, create_test_objects
from musicscore.tuplet import Tuplet
from musicscore.util import (
    XML_ARTICULATION_CLASSES,
    XML_TECHNICAL_CLASSES,
    XML_DYNAMIC_CLASSES,
    XML_ORNAMENT_CLASSES,
    XML_OTHER_NOTATIONS,
    XML_DIRECTION_TYPE_CLASSES,
)
from musicxml.xmlelement.xmlelement import *


def get_chord_midi_values(chord):
    return [m.value for m in chord._midis]


class TestTreeChord(ChordTestCase):
    def test_mocks(self):
        assert self.mock_voice.number == 1
        assert self.mock_beat.up.number == 1
        ch = Chord(60, 1)
        ch._parent = self.mock_beat
        assert ch.up == self.mock_beat
        assert ch.up.up == self.mock_voice
        assert ch.up.up.up == self.mock_staff
        assert ch.up.up.up.up == self.mock_measure
        assert ch.get_parent_measure() == self.mock_measure

    def test_chord_init_midis(self):
        """
        Test all possible types of input midis
        """
        ch = Chord(70, 1)
        assert get_chord_midi_values(ch) == [70]
        ch = Chord([70, 50], 1)
        assert get_chord_midi_values(ch) == [50, 70]
        ch = Chord(0, 1)
        assert get_chord_midi_values(ch) == [0]
        ch = Chord(Midi(90, accidental=Accidental(mode="enharmonic")), 1)
        assert get_chord_midi_values(ch) == [90]
        ch = Chord([Midi(90, accidental=Accidental(mode="enharmonic")), 70], 1)
        assert get_chord_midi_values(ch) == [70, 90]

        with self.assertRaises(ValueError):
            Chord([0, 60], 1)

        """
        A grace note cannot be a rest.
        """
        with self.assertRaises(ValueError):
            Chord(0, 0)

    def test_chord_needs_parent_error(self):
        ch = Chord(70, 1)
        with self.assertRaises(ChordHasNoParentBeamError):
            ch.finalize()
        ch._parent = self.mock_beat
        ch.finalize()

    def test_init_quarter_durations(self):
        """
        Test values of quarter_duration
        """
        ch = Chord(90, 1.25)
        assert ch.quarter_duration == 1.25
        ch = Chord(80, 1.2)
        assert ch.quarter_duration == Fraction(1.2)
        assert ch.quarter_duration == 1.2
        assert ch.quarter_duration == Fraction(6 / 5)
        assert ch.quarter_duration == 6 / 5
        ch = Chord(80, 8)
        assert ch.quarter_duration == 8
        with self.assertRaises(TypeError):
            ch = Chord(80, "1.2")
        ch = Chord(70, Fraction(1, 4))
        assert ch.quarter_duration == 0.25
        ch = Chord(70, QuarterDuration(1, 4))
        assert ch.quarter_duration == 0.25

    def test_is_rest(self):
        """
        Test is_rest property
        """
        assert Chord(0, 1).is_rest
        assert not Chord(50, 1).is_rest

    def test_chord_attributes(self):
        """
        Test that a dot operator can set and get note attributes
        """
        c = Chord([60, 61, 62], 2)
        c._parent = self.mock_beat
        c.finalize()
        c.relative_x = 10
        assert c.relative_x == [10, 10, 10]
        c.relative_y = [None, 20, 10]
        assert c.relative_y == [None, 20, 10]
        c.relative_y = [10]
        assert c.relative_y == [10, 20, 10]
        c.default_y = [10]
        assert c.default_y == [10, None, None]

    def test_chord_one_note(self):
        c = Chord(70, 4, relative_x=10)
        c.midis[0].accidental.show = True
        c._parent = self.mock_beat
        c.type = "whole"
        c.finalize()
        expected = """<note relative-x="10">
  <pitch>
    <step>B</step>
    <alter>-1</alter>
    <octave>4</octave>
  </pitch>
  <duration>4</duration>
  <voice>1</voice>
  <type>whole</type>
  <accidental>flat</accidental>
</note>
"""
        assert c.notes[0].to_string() == expected
        c.midis[0].value = 72
        c.type = "whole"
        expected = """<note relative-x="10">
  <pitch>
    <step>C</step>
    <octave>5</octave>
  </pitch>
  <duration>4</duration>
  <voice>1</voice>
  <type>whole</type>
  <accidental>natural</accidental>
</note>
"""
        assert c.notes[0].to_string() == expected
        # change chord's midi (zero)
        c.midis[0].value = 0
        # c.finalize()

        expected = """<note relative-x="10">
  <rest />
  <duration>4</duration>
  <voice>1</voice>
  <type>whole</type>
</note>
"""
        assert c.notes[0].to_string() == expected
        c = Chord(midis=0, quarter_duration=2)
        # change chord's duration (not zero)
        c.quarter_duration = 1
        c._parent = self.mock_beat
        c.type = "quarter"
        c.finalize()
        expected = """<note>
  <rest />
  <duration>1</duration>
  <voice>1</voice>
  <type>quarter</type>
</note>
"""
        assert c.notes[0].to_string() == expected
        c = Chord(0, 1)
        # change chord's duration (zero)
        with self.assertRaises(ValueError):
            c.quarter_duration = 0
        c.midis[0].value = 60
        # c.quarter_duration = 0
        # c._parent = self.mock_beat

    def test_chord_single_non_rest(self):
        """
        Test a chord with a non rest single midi
        """
        c = Chord(72, 2)
        c.midis[0].accidental.show = True
        c._parent = self.mock_beat
        c.finalize()

        c.xml_type = "16th"
        c.xml_stem = "up"
        c.xml_staff = 1

        expected = """<note>
  <pitch>
    <step>C</step>
    <octave>5</octave>
  </pitch>
  <duration>2</duration>
  <voice>1</voice>
  <type>16th</type>
  <accidental>natural</accidental>
  <stem>up</stem>
  <staff>1</staff>
</note>
"""
        assert c.notes[0].to_string() == expected
        c.midis[0].value = 61
        expected = """<pitch>
  <step>C</step>
  <alter>1</alter>
  <octave>4</octave>
</pitch>
"""
        assert c.notes[0].xml_pitch.to_string() == expected
        c.midis[0].value = 0
        # c.finalize()
        assert c.notes[0].xml_pitch is None
        assert c.notes[0].xml_rest is not None
        assert c.is_rest

    def test_chord_single_non_rest_midi_with_accidental(self):
        """
        Test chord with a non rest single midi with accidental set.
        """
        chord = Chord(Midi(70, accidental=Accidental(mode="sharp")), 1)
        chord._parent = self.mock_beat
        chord.finalize()
        expected = """<pitch>
    <step>A</step>
    <alter>1</alter>
    <octave>4</octave>
  </pitch>
"""
        assert chord.notes[0].xml_pitch.to_string() == expected

    def test_chord_as_rest(self):
        """
        Test chord with a rest midi
        """
        chord = Chord(0, 2)
        chord._parent = self.mock_beat
        chord.type = "half"
        expected = """<note>
  <rest />
  <duration>2</duration>
  <voice>1</voice>
  <type>half</type>
</note>
"""

        chord.finalize()
        assert chord.notes[0].to_string() == expected
        chord.midis = [60, 61]
        expected = """<pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
"""
        assert chord.notes[0].xml_pitch.to_string() == expected
        expected = """<pitch>
    <step>C</step>
    <alter>1</alter>
    <octave>4</octave>
  </pitch>
"""
        assert chord.notes[1].xml_pitch.to_string() == expected

    def test_chord_to_rest(self):
        chord = Chord(60, 2)
        chord._parent = self.mock_beat
        chord.type = chord.quarter_duration.get_type()
        chord.finalize()
        chord.to_rest()
        expected = """<note>
  <rest />
  <duration>2</duration>
  <voice>1</voice>
  <type>half</type>
</note>
"""
        assert chord.notes[0].to_string() == expected
        assert len(chord.notes) == 1

    def test_chord_with_multiple_midis(self):
        """
        Test chord with a list of midis
        """

        chord = Chord([60, 62, 63], 2)
        chord.midis[1].accidental.show = True
        chord.midis[2].accidental.show = True
        chord._parent = self.mock_beat
        chord.type = chord.quarter_duration.get_type()
        chord.finalize()
        chord.xml_stem = "up"
        expected_1 = """<note>
  <pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
  <duration>2</duration>
  <voice>1</voice>
  <type>half</type>
  <stem>up</stem>
</note>
"""
        expected_2 = """<note>
  <chord />
  <pitch>
    <step>D</step>
    <octave>4</octave>
  </pitch>
  <duration>2</duration>
  <voice>1</voice>
  <type>half</type>
  <accidental>natural</accidental>
  <stem>up</stem>
</note>
"""
        expected_3 = """<note>
  <chord />
  <pitch>
    <step>E</step>
    <alter>-1</alter>
    <octave>4</octave>
  </pitch>
  <duration>2</duration>
  <voice>1</voice>
  <type>half</type>
  <accidental>flat</accidental>
  <stem>up</stem>
</note>
"""
        for note, expected in zip(chord.notes, [expected_1, expected_2, expected_3]):
            assert note.to_string() == expected

    def test_group_chords(self):
        chords = [Chord(60, qd) for qd in [1 / 6, 1 / 6, 1 / 6, 1 / 10, 3 / 10, 1 / 10]]
        with self.assertRaises(ValueError):
            assert _group_chords(chords, [1 / 2, 1 / 2, 1 / 2])
        assert _group_chords(chords, [1 / 2, 1 / 2]) == [chords[:3], chords[3:]]
        assert _group_chords(chords, [1 / 3, 2 / 3]) == [chords[:2], chords[2:]]
        assert _group_chords(chords, [1 / 4, 3 / 4]) is None

    def test_has_same_pitches(self):
        ch1 = Chord([60, Midi(61), 62], 1)
        ch2 = Chord([60, Midi(61)], 1)
        assert not ch1.has_same_pitches(ch2)
        ch2 = Chord([60, Midi(61), 62], 1)
        assert ch1.has_same_pitches(ch2)
        ch2 = Chord([60, Midi(61, accidental=Accidental(mode="flat")), 62], 1)
        assert not ch1.has_same_pitches(ch2)

    def test_add_lyric(self):
        ch = Chord(60, 2)
        ch._parent = self.mock_beat
        ch.add_lyric("test")
        ch.finalize()
        assert ch.notes[0].xml_lyric is not None
        assert ch.notes[0].xml_lyric.xml_text.value_ == "test"

    def test_get_staff_number(self):
        ch = Chord(60, 2)
        ch._parent = self.mock_beat
        assert ch.get_staff_number() is None
        self.mock_staff.number = 1
        assert ch.get_staff_number() == 1

    def test_add_clef(self):
        ch = Chord(60, 2)
        assert ch.clef is None
        cl = BassClef()
        ch.clef = cl
        assert ch.clef == cl
        with self.assertRaises(TypeError):
            ch.clef = "bla"

    def test_break_beam(self):
        ch = Chord(60, 2)
        self.assertFalse(ch.broken_beam)
        ch.broken_beam = True
        self.assertTrue(ch.broken_beam)
        ch = Chord(60, 2)
        self.assertFalse(ch.broken_beam)
        ch.break_beam()
        self.assertTrue(ch.broken_beam)

    @skip
    def test_percussion_notation(self):
        self.fail("Incomplete")

    @skip
    def test_finger_tremolo(self):
        self.fail("Incomplete")

    def test_deepcopy_chord(self):
        chord = Chord([60, 62], 2)
        chord.add_tie("start")
        copied = copy.deepcopy(chord)
        assert [midi.value for midi in copied.midis] == [
            midi.value for midi in chord.midis
        ]
        assert [id(midi) for midi in copied.midis] != [id(midi) for midi in chord.midis]
        assert chord.quarter_duration.value == copied.quarter_duration.value
        assert id(chord.quarter_duration) != id(copied.quarter_duration)
        chord._parent = self.mock_beat
        chord.finalize()
        with self.assertRaises(DeepCopyException):
            copy.deepcopy(chord)
        # chord.add_dynamics('p')
        # chord.add_lyric('something')
        # chord.add_x(XMLAccent())
        # chord.add_x(XMLUpBow())
        # chord.add_x(XMLAccidentalMark())
        # chord.add_x(XMLFf())

    def test_add_midi(self):
        chord = Chord([60, 62], 2)
        m = Midi(63)
        chord.add_midi(m)
        assert [midi.value for midi in chord.midis] == [60, 62, 63]
        assert chord.midis[-1] == m
        chord.add_midi(58)
        assert [midi.value for midi in chord.midis] == [58, 60, 62, 63]
        chord.add_midi(60)
        assert [midi.value for midi in chord.midis] == [58, 60, 60, 62, 63]

        chord._parent = self.mock_beat
        chord.finalize()
        with self.assertRaises(AlreadyFinalizedError):
            chord.add_midi(80)

    def test_add_direction_type(self):
        score = Score()
        p = score.add_part("part-1")
        for dt_class in XML_DIRECTION_TYPE_CLASSES:
            chord = Chord(midis=60, quarter_duration=4)
            if dt_class == XMLSymbol:
                dt_obj = dt_class("0")
            else:
                dt_obj = dt_class()
            if dt_class == XMLDynamics:
                assert self.assertRaises(ChordException)
            else:
                chord.add_direction_type(dt_obj, "above")
                assert dt_obj in chord.xml_direction_types["above"]
                assert chord.xml_direction_types["above"] == [dt_obj]
                p.add_chord(chord)
        p.finalize()

    def test_add_direction_type_wrong_type(self):
        chord = Chord(60, 4)
        with self.assertRaises(TypeError):
            chord.add_direction_type(XMLFermata())

    def test_chord_to_string(self):
        p = Part("p1")
        ch = Chord(60, 1)
        p.add_chord(ch)
        ch.finalize()
        with self.assertRaises(MusicTreeException):
            ch.to_string()

    def test_get_x(self):
        def test_get_objects(*classes):
            chord = Chord(60, 1)
            cl1, cl2, cl3 = classes
            dt1 = chord.add_x(cl1())
            dt2 = chord.add_x(cl1())
            dt3 = chord.add_x(cl2())
            assert chord.get_x(cl1) == [dt1, dt2]
            assert chord.get_x(cl2) == [dt3]
            assert chord.get_x(cl3) == []

        # oranment
        test_get_objects(
            XML_ORNAMENT_CLASSES[0], XML_ORNAMENT_CLASSES[1], XML_ORNAMENT_CLASSES[2]
        )
        # articualtion
        test_get_objects(
            XML_ARTICULATION_CLASSES[0],
            XML_ARTICULATION_CLASSES[1],
            XML_ARTICULATION_CLASSES[2],
        )
        # technical
        test_get_objects(
            XML_TECHNICAL_CLASSES[0], XML_TECHNICAL_CLASSES[1], XML_TECHNICAL_CLASSES[2]
        )
        # other notations
        test_get_objects(
            XML_OTHER_NOTATIONS[0], XML_OTHER_NOTATIONS[1], XML_OTHER_NOTATIONS[2]
        )
        # direction type
        test_get_objects(
            XML_DIRECTION_TYPE_CLASSES[0],
            XML_DIRECTION_TYPE_CLASSES[1],
            XML_DIRECTION_TYPE_CLASSES[2],
        )

        # ornament or notation
        chord = Chord(60, 1)
        cl1 = XMLAccidentalMark
        dt1 = chord.add_x(cl1("sharp"), parent_type="ornament")
        dt2 = chord.add_x(cl1("flat"), parent_type="ornament")
        dt3 = chord.add_x(cl1("flat"), parent_type="notation")
        assert chord.get_x(cl1) == [dt1, dt2, dt3]

    def test_get_wedges(self):
        chord = Chord(60, 1)
        dt1 = chord.add_x(XMLWedge(type="crescendo"))
        dt2 = chord.add_x(XMLWedge(type="continue"))
        dt3 = chord.add_x(XMLWedge(type="stop"))
        assert chord.get_wedges() == [dt1, dt2, dt3]

    def test_get_slurs(self):
        chord = Chord(60, 1)
        dt1 = chord.add_x(XMLSlur(type="start"))
        dt2 = chord.add_x(XMLSlur(type="continue"))
        dt3 = chord.add_x(XMLSlur(type="stop"))
        assert chord.get_slurs() == [dt1, dt2, dt3]

    def test_get_words(self):
        chord = Chord(60, 1)
        dt1 = chord.add_x(XMLWords("something"))
        dt2 = chord.add_x(XMLWords("something else"))
        assert chord.get_words() == [dt1, dt2]

    def test_get_brackets(self):
        chords = [Chord(60, 1) for _ in range(3)]
        b1, b2, b3 = (
            chords[0].add_x(XMLBracket()),
            chords[1].add_x(XMLBracket()),
            chords[2].add_x(XMLBracket()),
        )
        assert [ch.get_brackets()[0] for ch in chords] == [b1, b2, b3]

    def test_add_arpeggio_normal(self):
        ch = Chord([60, 64, 67], 1)
        assert not ch.arpeggio
        ch._parent = self.mock_beat
        ch.arpeggio = "normal"
        assert ch.arpeggio == "normal"
        ch.finalize()
        for n in ch.notes:
            assert n.xml_notations.xml_arpeggiate
            assert n.xml_notations.xml_arpeggiate.direction is None

    def test_add_arpeggio_up(self):
        ch = Chord([60, 64, 67], 1)
        assert not ch.arpeggio
        ch._parent = self.mock_beat
        ch.arpeggio = "up"
        assert ch.arpeggio == "up"
        ch.finalize()
        for n in ch.notes:
            assert n.xml_notations.xml_arpeggiate
            if n == ch.notes[0]:
                assert n.xml_notations.xml_arpeggiate.direction == "up"

    def test_add_arpeggio_bottom(self):
        ch = Chord([60, 64, 67], 1)
        assert not ch.arpeggio
        ch._parent = self.mock_beat
        ch.arpeggio = "down"
        assert ch.arpeggio == "down"
        ch.finalize()
        for n in ch.notes:
            assert n.xml_notations.xml_arpeggiate
            if n == ch.notes[0]:
                assert n.xml_notations.xml_arpeggiate.direction == "down"

    def test_add_none_arpeggio(self):
        ch = Chord([60, 64, 67], 1)
        assert not ch.arpeggio
        ch._parent = self.mock_beat
        ch.arpeggio = "none"
        assert ch.arpeggio == "none"
        ch.finalize()
        for n in ch.notes:
            if n == ch.notes[0]:
                assert n.xml_notations.xml_non_arpeggiate
                assert n.xml_notations.xml_non_arpeggiate.type == "bottom"
            elif n == ch.notes[-1]:
                assert n.xml_notations.xml_non_arpeggiate
                assert n.xml_notations.xml_non_arpeggiate.type == "top"


class TestTreeRest(ChordTestCase):
    def test_rest_init(self):
        r = Rest(4)
        r._parent = self.mock_beat
        assert isinstance(r, Rest)
        assert isinstance(r, Chord)
        assert r.midis[0].value == 0
        r.finalize()
        n = r.notes[0]
        assert (
            r.display_step
            == r.display_octave
            == n.xml_object.xml_rest.xml_display_step
            == n.xml_object.xml_rest.xml_display_octave
            is None
        )
        assert r.quarter_duration == 4
        r = Rest(4, display_step="C", display_octave=4)
        r._parent = self.mock_beat
        r.finalize()
        n = r.notes[0]
        assert r.display_step == n.xml_object.xml_rest.xml_display_step.value_ == "C"
        assert r.display_octave == n.xml_object.xml_rest.xml_display_octave.value_ == 4
        r = Rest(4, measure="yes")
        assert r.measure == "yes"
        r._parent = self.mock_beat
        r.finalize()
        n = r.notes[0]
        assert n.xml_object.xml_rest.measure == "yes"
        with self.assertRaises(RestCannotSetMidiError):
            Rest(midis=0, quarter_duration=2)
        with self.assertRaises(TypeError):
            Rest()
        with self.assertRaises(RestWithDisplayStepHasNoDisplayOctave):
            r = Rest(4, display_step="D")
            r._parent = self.mock_beat
            r.finalize()
        with self.assertRaises(RestWithDisplayOctaveHasNoDisplayStep):
            r = Rest(4, display_octave=4)
            r._parent = self.mock_beat
            r.finalize()
        with self.assertRaises(TypeError):
            Rest(4, display_step="H", display_octave=2)
        with self.assertRaises(TypeError):
            Rest(4, display_step="C", display_octave=2.2)
        with self.assertRaises(TypeError):
            Rest(4, display_step="C", display_octave=-2)


class TestTies(ChordTestCase):
    def test_add_tie_ties_midis(self):
        ch = Chord([60, 63], 1)
        ch._parent = self.mock_beat
        ch.add_tie("start")
        for midi in ch.midis:
            assert midi._ties == {"start"}
        ch.finalize()
        assert [n.is_tied for n in ch.notes] == [True, True]

    def test_tied_midis(self):
        m1 = Midi(60)
        m1.add_tie("start")
        m2 = Midi(61)
        m2.add_tie("start")
        ch = Chord([m1, m2], 1)
        ch._parent = self.mock_beat
        ch.finalize()
        assert [n.is_tied for n in ch.notes] == [True, True]

    def test_split_tied_copy(self):
        ch = Chord(midis=60, quarter_duration=1)
        copied = _split_copy(ch)
        assert ch.midis[0]._ties == copied.midis[0]._ties == set()

    def test_tie_one_note(self):
        ch1 = Chord([60, 63], 1)
        ch2 = Chord([60, 65], 1)
        ch1.midis[0].add_tie("start")
        ch2.midis[0].add_tie("stop")
        ch1._parent = self.mock_beat
        ch2._parent = self.mock_beat
        ch1.finalize()
        ch2.finalize()
        assert [n.is_tied for n in ch1.notes] == [True, False]
        assert [n.is_tied_to_previous for n in ch2.notes] == [True, False]

    def test_untie_one_note(self):
        ch = Chord([60, 61], 1)
        ch.add_tie("start")
        ch._parent = self.mock_beat
        ch.finalize()
        assert [n.is_tied for n in ch.notes] == [True, True]
        # print(ch.notes)
        ch.midis[0].remove_tie("start")
        assert [n.is_tied for n in ch.notes] == [False, True]

    def test_chord_tie_untie(self):
        ch1 = Chord([60, 61], 1)
        ch2 = Chord([60, 61], 1)
        ch1.add_tie("start")
        ch2.add_tie("stop")
        ch1._parent = self.mock_beat
        ch2._parent = self.mock_beat
        ch1.finalize()
        ch2.finalize()
        assert [n.is_tied for n in ch1.notes] == [True, True]
        assert [n.is_tied for n in ch2.notes] == [False, False]
        assert [n.is_tied_to_previous for n in ch2.notes] == [True, True]

    def test_chord_change_tie_after_finalizing(self):
        ch1 = Chord([60, 61], 1)
        ch2 = Chord([60, 61], 1)
        ch1._parent = self.mock_beat
        ch2._parent = self.mock_beat
        ch1.finalize()
        ch2.finalize()

        ch1.add_tie("start")
        ch2.midis[1].add_tie("stop")

        assert [n.is_tied for n in ch1.notes] == [True, True]
        assert [n.is_tied for n in ch2.notes] == [False, False]
        assert [n.is_tied_to_previous for n in ch2.notes] == [False, True]

    def test_chord_all_midis_tied_to_next_or_previous(self):
        ch = Chord([60, 61, 62], 1)
        for m in ch.midis:
            m.add_tie("start")
        assert ch.all_midis_are_tied_to_next
        ch.midis[1].remove_tie("start")
        assert not ch.all_midis_are_tied_to_next
        ch.midis[0].add_tie("stop")
        assert not ch.all_midis_are_tied_to_previous
        for m in ch.midis:
            m.add_tie("stop")
        assert ch.all_midis_are_tied_to_previous


class TestSplit(TestCase):
    def test_set_original_starting_ties(self):
        ch = Chord(midis=[60, 61], quarter_duration=1)
        ch._set_original_starting_ties(ch)
        copied = _split_copy(ch)
        assert copied._original_starting_ties == [set(), set()]
        ch.midis[0].add_tie("start")
        ch._set_original_starting_ties(ch)
        assert copied._original_starting_ties == [set(), set()]

    def test_split_copy(self):
        ch = Chord(
            midis=[Midi(61, accidental=Accidental(mode="sharp"))],
            quarter_duration=2,
            offset=0.5,
        )
        copied = _split_copy(ch)
        assert ch.midis != copied.midis
        assert ch.midis[0].value == copied.midis[0].value
        assert ch.midis[0].accidental != copied.midis[0].accidental
        assert ch.midis[0].accidental.mode == copied.midis[0].accidental.mode
        copied.midis[0].accidental.show = False
        assert ch.midis[0].accidental.show is None
        assert copied.midis[0].accidental.show is False

    def test_tied_split_copy(self):
        ch = Chord(midis=61, quarter_duration=2)
        ch.add_tie("start")
        copied = _split_copy(ch)
        for m in copied.midis:
            assert m._ties == set()

    def test_split_quarter_durations(self):
        ch = Chord(midis=60, quarter_duration=4)
        copied = _split_copy(ch)
        assert id(ch.quarter_duration) != id(copied.quarter_duration)
        ch.quarter_duration = 2
        assert copied.quarter_duration == 4
        assert ch.quarter_duration == 2

        ch = Chord(midis=60, quarter_duration=5)
        beats = [Beat(1), Beat(1), Beat(1), Beat(1)]
        quarter_durations = ch.quarter_duration._get_beatwise_sections(beats=beats)
        ch.quarter_duration = quarter_durations[0][0]
        copied = _split_copy(ch, quarter_durations[1])
        assert [ch.quarter_duration, copied.quarter_duration] == [4, 1]


class TestAddGraceChord(ChordTestCase):
    def test_add_grace_chord_parameters(self):
        ch = Chord(60, 1)
        g1 = gch = ch.add_grace_chord(midis_or_grace_chord=60)
        assert gch.midis[0].value == 60
        in_gch = GraceChord(61)
        g2 = gch = ch.add_grace_chord(midis_or_grace_chord=in_gch)
        assert gch == in_gch
        g3 = gch = ch.add_grace_chord(midis_or_grace_chord=[62, 63])
        assert [m.value for m in gch.midis] == [62, 63]
        g4 = gch = ch.add_grace_chord(64)
        assert gch.midis[0].value == 64
        g5 = gch = ch.add_grace_chord(midis_or_grace_chord=65, type="16th")
        assert gch.midis[0].value == 65
        assert gch.type == "16th"
        in_gch = GraceChord(61, type="16th")
        g6 = gch = ch.add_grace_chord(midis_or_grace_chord=in_gch)
        assert gch == in_gch
        assert gch.type == "16th"
        with self.assertRaises(ValueError):
            ch.add_grace_chord(
                midis_or_grace_chord=GraceChord(66, type="quarter"), type="16th"
            )
        g7 = gch = ch.add_grace_chord(67, type="16th")
        assert gch.midis[0].value == 67
        assert gch.type == "16th"
        g8 = gch = ch.add_grace_chord(68, type="16th", position="after")
        assert gch.position == "after"
        assert gch.midis[0].value == 68
        assert gch.type == "16th"
        with self.assertRaises(ValueError):
            ch.add_grace_chord(GraceChord(66, type="quarter"), position="after")
        assert ch.get_grace_chords(position="before") == [g1, g2, g3, g4, g5, g6, g7]
        assert ch.get_grace_chords(position="after") == [g8]

    def test_add_grace_chord_before_and_after_argument(self):
        ch = Chord(60, 1)
        midis_or_chord = [62, [63, 65], GraceChord(66)]
        before_chords = [ch.add_grace_chord(x) for x in midis_or_chord]
        assert ch._grace_chords["before"] == before_chords

        midis_or_chord = [67, [68, 70], GraceChord(71, position="after")]
        after_chords = [
            ch.add_grace_chord(x, position="after")
            if not isinstance(x, GraceChord)
            else ch.add_grace_chord(x)
            for x in midis_or_chord
        ]
        assert ch._grace_chords["after"] == after_chords

    def test_add_grace_chord_generates_grace_chords(self):
        ch = Chord(60, 1)
        with self.assertRaises(TypeError):
            ch.add_grace_chord()
        gc1 = ch.add_grace_chord(60)
        gc2 = ch.add_grace_chord(61, type="quarter")
        gc3 = ch.add_grace_chord(GraceChord(midis=[62, 64]))
        gc4 = ch.add_grace_chord(GraceChord(63, type="16th"))
        all_gcs = [gc1, gc2, gc3, gc4]
        assert [[m.value for m in gc.midis] for gc in all_gcs] == [
            [60],
            [61],
            [62, 64],
            [63],
        ]
        assert [gc.type for gc in all_gcs] == [None, "quarter", None, "16th"]

    def test_add_grace_chord_finalize(self):
        part = Part("p1")
        ch = Chord(midis=[60, 63], quarter_duration=4)
        midis_or_chord = [62, [63, 65], GraceChord(66)]
        # before_chords
        [ch.add_grace_chord(x) for x in midis_or_chord]
        midis_or_chord = [67, [68, 70], GraceChord(71, position="after")]
        # after chords
        [
            ch.add_grace_chord(x, position="after")
            if not isinstance(x, GraceChord)
            else ch.add_grace_chord(x)
            for x in midis_or_chord
        ]
        part.add_chord(ch)
        with self.assertRaises(ChordException):
            ch.add_grace_chord(80)
        part.finalize()
        assert len(part.get_chords()) == 7

    def test_grace_chords_are_added_to_right_voice(self):
        part = Part("p1")
        ch = Chord(midis=[60, 63], quarter_duration=4)
        gch1 = ch.add_grace_chord(80)
        gch2 = ch.add_grace_chord(90, position="after")
        part.add_chord(ch, staff_number=2, voice_number=2)
        part.finalize()
        assert part.get_voice(1, 2, 2).get_chords() == [gch1, ch, gch2]
        for c in [gch1, ch, gch2]:
            assert c.voice_number == 2
            assert c.get_staff_number() == 2

    def test_grace_chords_after_the_last_chord_in_measure(self):
        part = Part("p1")
        ch = Chord(60, 4)
        gch = ch.add_grace_chord(80, position="after")
        part.add_chord(ch)
        part.add_chord(Chord(90, 1))
        part.finalize()
        assert gch in part.get_measure(1).get_chords()

    def test_error_grace_chord_add_grace_chord(self):
        gch = GraceChord(60)
        with self.assertRaises(GraceChordCannotHaveGraceNotesError):
            gch.add_grace_chord(80)

    def test_grace_chords_after_to_right_beat(self):
        part = Part("p1")
        ch = Chord(60, 1)
        gch = ch.add_grace_chord(80, position="after")
        part.add_chord(ch)
        part.add_chord(Chord(90, 3))
        assert gch in part.get_beat(1, 1, 1, 1).get_children()


class TestAddX(ChordTestCase):
    # def setUp(self):
    #     super().setUp()
    #     self.score = Score()
    #     self.part = self.score.add_part('p1')

    def test_add_x_parent_type(self):
        ch = Chord(60, 1)
        ch.add_x(XML_DIRECTION_TYPE_CLASSES[0]())

    def test_add_x_placement(self):
        def add_type_and_placement(type, xml_object=None):
            above_part = (
                Part(f"above-{type}-{xml_object.name}")
                if xml_object
                else Part(f"above-{type}")
            )
            below_part = (
                Part(f"below-{type}-{xml_object.name}")
                if xml_object
                else Part(f"below-{type}")
            )
            if not xml_object:
                above_objects = create_test_objects(type)
                below_objects = create_test_objects(type)
            else:
                above_objects = [xml_object]
                below_objects = [xml_object.__deepcopy__()]

            for obj in above_objects:
                ch = Chord(60, 1)
                try:
                    ch.add_x(obj, placement="above", parent_type=type)
                except ChordAddXPlacementException:
                    ch.add_x(obj, placement=None, parent_type=type)
                above_part.add_chord(ch)
            for obj in below_objects:
                ch = Chord(60, 1)
                try:
                    ch.add_x(obj, placement="below", parent_type=type)
                except ChordAddXPlacementException:
                    ch.add_x(obj, placement=None, parent_type=type)
                below_part.add_chord(ch)

            above_part.finalize()
            below_part.finalize()
            if type == "direction_type":
                for p in ["above", "below"]:
                    directions = [
                        d
                        for m in eval(f"{p}_part.get_children()")
                        for d in [
                            c
                            for c in m.xml_object.get_children()
                            if isinstance(c, XMLDirection)
                        ]
                    ]
                    assert len(directions) == len(above_objects)
                    for x in directions:
                        assert x.placement == p

            elif type in ["technical", "ornament", "articulation"]:
                for p in ["above", "below"]:
                    notations = [
                        notation
                        for measure in eval(f"{p}_part").get_children()
                        for note in [
                            x
                            for x in measure.xml_object.get_children()
                            if isinstance(x, XMLNote)
                        ]
                        for notation in [
                            y
                            for y in note.get_children()
                            if isinstance(y, XMLNotations)
                        ]
                    ]
                    assert len(notations) == len(above_objects)
                    for x in notations:
                        if x.get_children()[0].get_children()[0].__class__ not in [
                            XMLFret,
                            XMLBend,
                        ]:
                            assert x.get_children()[0].get_children()[0].placement == p
            else:
                raise NotImplementedError(f"testing type {type}")

        # direction_type objects
        add_type_and_placement("direction_type")

        # technical objects
        add_type_and_placement("technical")

        # ornament objects
        add_type_and_placement("ornament")

        # articulation objects
        add_type_and_placement("articulation")

        # notation only objects does not accept placement (except fermata)
        notation_object = XMLArpeggiate()
        ch = Chord(60, 1)
        with self.assertRaises(ChordAddXPlacementException):
            ch.add_x(notation_object, placement="above", parent_type="notation")
        ch.add_x(notation_object, placement=None, parent_type="notation")

        # mixed below and above
        part = Part("mixed_placements")
        x_objects = [XMLTenuto(), XMLStaccato(), XMLAccent()]
        placements = ["below", "below", "above"]
        ch = Chord(60, 1)
        for obj, plc in zip(x_objects, placements):
            ch.add_x(obj, placement=plc)
        part.add_chord(ch)
        part.finalize()
        notations = [
            notation
            for measure in part.get_children()
            for note in [
                x for x in measure.xml_object.get_children() if isinstance(x, XMLNote)
            ]
            for notation in [
                y for y in note.get_children() if isinstance(y, XMLNotations)
            ]
        ]
        for index, x in enumerate(notations[0].xml_articulations.get_children()):
            assert x.placement == placements[index]

        # fermata placement is translated in fermata types inverted or upright
        fermatas = [XMLFermata(), XMLFermata()]
        placements = ["above", "below"]
        ch = Chord(60, 1)
        for f, p in zip(fermatas, placements):
            ch.add_x(f, placement=p)

        for f, t in zip(fermatas, ["upright", "inverted"]):
            assert f.type == t

    def test_add_articulation_after_creating_notes(self):
        ch = Chord(60, 1)
        ch._parent = self.mock_beat
        staccato = ch.add_x(XMLStaccato())
        ch.finalize()
        assert ch.notes[0].xml_notations.xml_articulations.get_children() == [staccato]
        assert isinstance(
            ch.notes[0].xml_notations.xml_articulations.get_children()[0], XMLStaccato
        )
        accent = ch.add_x(XMLAccent())
        assert ch.notes[0].xml_notations.xml_articulations.get_children() == [
            staccato,
            accent,
        ]

    def test_add_multiple_articulations(self):
        articulation_classes = XML_ARTICULATION_CLASSES[:3]
        ch = Chord(60, 1)
        ch._parent = self.mock_beat
        for a in articulation_classes:
            ch.add_x(a())
        assert len(ch._xml_articulations) == 3
        ch.finalize()
        n = ch.notes[0]
        assert n.xml_notations.xml_articulations is not None
        assert [
            type(a) for a in n.xml_notations.xml_articulations.get_children()
        ] == articulation_classes

    def test_chord_add_x_as_object_articulation(self):
        for obj in create_test_objects("articulation"):
            ch = Chord(60, 1)
            ch.add_x(obj)
            ch._parent = self.mock_beat
            ch.finalize()
            assert ch.notes[0].xml_notations.xml_articulations.get_children()[0] == obj

    def test_chord_add_x_as_object_technical(self):
        for obj in create_test_objects("technical"):
            ch = Chord(60, 1)
            ch.add_x(obj)
            ch._parent = self.mock_beat
            ch.finalize()
            assert ch.notes[0].xml_notations.xml_technical.get_children()[0] == obj

    def test_chord_add_x_as_object_dynamics(self):
        for cls in XML_DYNAMIC_CLASSES:
            ch = Chord(60, 1)
            ch.add_x(cls(), parent_type="notation")
            ch._parent = self.mock_beat
            ch.finalize()
            assert isinstance(
                ch.notes[0].xml_notations.xml_dynamics.get_children()[0], cls
            )

    def test_chord_add_x_as_object_ornaments(self):
        for obj in create_test_objects("ornament"):
            ch = Chord(60, 1)
            if isinstance(obj, XMLAccidentalMark):
                ch.add_x(obj, parent_type="ornament")
            else:
                ch.add_x(obj)
            ch._parent = self.mock_beat
            ch.finalize()
            assert ch.notes[0].xml_notations.xml_ornaments.get_children()[0] == obj

    def test_chord_add_x_trill_with_wavy_line_and_accidental_mark(self):
        ch = Chord(60, 1)
        ch.add_x(XMLTrillMark())
        ch.add_x(XMLAccidentalMark("sharp"), parent_type="ornament")
        ch.add_x(XMLWavyLine(type="start", relative_x=0))
        ch.add_x(XMLWavyLine(type="stop", relative_x=20))
        ch._parent = self.mock_beat
        ch.finalize()
        expected = """<ornaments>
      <trill-mark />
      <accidental-mark>sharp</accidental-mark>
      <wavy-line type="start" relative-x="0" />
      <wavy-line type="stop" relative-x="20" />
    </ornaments>
"""
        assert ch.notes[0].xml_notations.xml_ornaments.to_string() == expected

    def test_chord_add_x_as_object_other_notations(self):
        for cls in XML_OTHER_NOTATIONS:
            ch = Chord(60, 1)
            ch.add_x(cls())
            ch._parent = self.mock_beat
            ch.finalize()
            assert isinstance(ch.notes[0].xml_notations.get_children()[0], cls)

    def test_add_xml_wedge_objects(self):
        wedges = [
            XMLWedge(type=val) for val in ["crescendo", "stop", "diminuendo", "stop"]
        ]
        for wedge in wedges:
            ch = Chord(60, 4)
            ch._parent = self.mock_beat
            ch.add_wedge(wedge)
            ch.finalize()
            assert len(ch._xml_directions) == 1
            d = ch._xml_directions[0]
            assert d.placement == "below"
            assert d.xml_direction_type.xml_wedge == wedge

    def test_add_wedge_string(self):
        wedges = ["crescendo", "stop", "diminuendo", "stop"]
        for wedge in wedges:
            ch = Chord(60, 4)
            ch._parent = self.mock_beat
            ch.add_wedge(wedge)
            ch.finalize()
            assert len(ch._xml_directions) == 1
            d = ch._xml_directions[0]
            assert d.placement == "below"
            assert d.xml_direction_type.xml_wedge.type == wedge

    def test_add_words(self):
        ch = Chord(60, 4)
        ch._parent = self.mock_beat
        texts = ["Below", "Above", "FontSize", "Bold"]
        xml_words_objects = [
            ch.add_words(texts[0], placement="below"),
            ch.add_words(texts[1]),
            ch.add_words(texts[2], font_size=18, relative_y=30),
            ch.add_words(texts[3], font_weight="bold", relative_y=60),
        ]
        ch.finalize()
        assert set(xml_words_objects) == set(
            [d.xml_direction_type.xml_words for d in ch._xml_directions]
        )

        for obj, t in zip(xml_words_objects, texts):
            assert obj.value_ == t

        for obj, p in zip(xml_words_objects, ["below", "above", "above", "above"]):
            assert obj.up.up.placement == p
        for fs, obj in zip([None, None, 18, None], xml_words_objects):
            assert obj.font_size == fs
        for fw, obj in zip([None, None, None, "bold"], xml_words_objects):
            assert obj.font_weight == fw
        for ry, obj in zip([None, None, 30, 60], xml_words_objects):
            assert obj.relative_y == ry

        ch = Chord(60, 4)
        ch._parent = self.mock_beat

        xml_words = ch.add_words(
            XMLWords("something", font_size=10, relative_y=20),
            placement="below",
            font_size=20,
        )
        ch.finalize()
        d = ch._xml_directions[0]
        assert d.xml_direction_type.xml_words.value_ == "something"
        assert d.placement == "below"
        assert xml_words.font_size == 20
        assert xml_words.relative_y == 20


class TestAddAfterNotes(TestCase):
    def test_add_after_notes(self):
        part = Part("p1")
        chords = [Chord([60, 62], 2), Chord(64, 2)]
        chords[0].add_dynamics("ff")
        [part.add_chord(ch) for ch in chords]
        b = XMLBarline(location="middle")
        b.xml_bar_style = "dashed"
        chords[0].add_xml_element_after_notes(b)
        part.finalize()
        m = part.get_measure(1)
        expected = """<measure number="1">
    <attributes>
      <divisions>1</divisions>
      <key>
        <fifths>0</fifths>
      </key>
      <time>
        <beats>4</beats>
        <beat-type>4</beat-type>
      </time>
      <clef>
        <sign>G</sign>
        <line>2</line>
      </clef>
    </attributes>
    <direction placement="below">
      <direction-type>
        <dynamics>
          <ff />
        </dynamics>
      </direction-type>
    </direction>
    <note>
      <pitch>
        <step>C</step>
        <octave>4</octave>
      </pitch>
      <duration>2</duration>
      <voice>1</voice>
      <type>half</type>
    </note>
    <note>
      <chord />
      <pitch>
        <step>D</step>
        <octave>4</octave>
      </pitch>
      <duration>2</duration>
      <voice>1</voice>
      <type>half</type>
    </note>
    <barline location="middle">
      <bar-style>dashed</bar-style>
    </barline>
    <note>
      <pitch>
        <step>E</step>
        <octave>4</octave>
      </pitch>
      <duration>2</duration>
      <voice>1</voice>
      <type>half</type>
    </note>
  </measure>
"""
        assert m.to_string() == expected


class TestTypeAndNumberOfDots(TestCase):
    def test_chord_type_error(self):
        ch = Chord(60, 1)
        with self.assertRaises(ValueError):
            ch.type = "bla"

    def test_chord_number_of_dots_error(self):
        ch = Chord(60, 1)
        with self.assertRaises(TypeError):
            ch.number_of_dots = None
        with self.assertRaises(TypeError):
            ch.number_of_dots = "1"
        with self.assertRaises(ValueError):
            ch.number_of_dots = -1

    def test_chord_type_attribute_with_update(self):
        p = Part("p0")
        ch = Chord([60, 61], 2)
        p.add_chord(ch)
        p.finalize()
        assert ch.get_children()[0].xml_type.value_ == "half"
        assert ch.get_children()[1].xml_type.value_ == "half"
        p = Part("p1")
        ch = Chord([60, 61], 2)
        ch.type = "whole"
        p.add_chord(ch)
        p.finalize()
        assert ch.get_children()[0].xml_type.value_ == "whole"
        assert ch.get_children()[1].xml_type.value_ == "whole"
        with self.assertRaises(ChordAlreadyHasNotesError):
            ch.type = "half"
        gch = Chord([60, 61], 0)
        assert gch.type is None
        p = Part("p2")
        p.add_chord(gch)
        # print(p.get_beats()[0])
        # print(p.get_beats()[0].is_filled)
        p.finalize()

        assert gch.get_children()[0].xml_type is None
        assert gch.get_children()[1].xml_type is None

        p = Part("p3")
        p.add_chord(Chord(60, 1))
        gch = GraceChord([60, 61])
        assert gch.type is None
        p.add_chord(gch)
        p.finalize()

        assert gch.get_children()[0].xml_type is None
        assert gch.get_children()[1].xml_type is None

        p = Part("p4")
        gch = GraceChord([60, 61])
        gch.type = "half"
        assert gch.type == "half"
        p.add_chord(gch)
        p.finalize()
        assert gch.get_children()[0].xml_type.value_ == "half"
        assert gch.get_children()[1].xml_type.value_ == "half"

    def test_chord_number_of_dots_with_update(self):
        p = Part("p1")
        [p.add_chord(Chord([60, 61], qd)) for qd in [1, 0.75, 0.25]]
        assert {ch.number_of_dots for ch in p.get_chords()} == {None}
        p.finalize()
        assert [ch.number_of_dots for ch in p.get_chords()[:3]] == [0, 1, 0]
        ch = p.get_chords()[1]
        assert len(ch.get_children()[0].xml_object.get_children_of_type(XMLDot)) == 1
        assert len(ch.get_children()[1].xml_object.get_children_of_type(XMLDot)) == 1
        with self.assertRaises(ChordAlreadyHasNotesError):
            ch.number_of_dots = 3

        p = Part("p2")
        gch = GraceChord([60, 61])
        gch.number_of_dots = 3
        p.add_chord(gch)
        p.finalize()
        assert len(gch.get_children()[0].xml_object.get_children_of_type(XMLDot)) == 3
        assert len(gch.get_children()[1].xml_object.get_children_of_type(XMLDot)) == 3

    def test_number_of_beams(self):
        chord = Chord(60, 1)
        with self.assertRaises(ChordTypeNotSetError):
            chord.number_of_beams
        chord.type = "quarter"
        assert chord.number_of_beams == 0
        chord.type = "32nd"
        assert chord.number_of_beams == 3

    def test_test_quarter_duration(self):
        chord = Chord(60, 1 / 3)
        with self.assertRaises(ChordHasNoParentBeamError):
            chord.check_printed_duration()
        chord._parent = Mock()
        chord._parent.quarter_duration = None
        with self.assertRaises(ChordParentBeamError):
            chord.check_printed_duration()
        chord._parent.quarter_duration = 1
        chord._parent.get_subdivision.return_value = None
        with self.assertRaises(ChordParentBeamError):
            chord.check_printed_duration()
        chord._parent.get_subdivision.return_value = 3
        with self.assertRaises(ChordTypeNotSetError):
            chord.check_printed_duration()
        chord.type = "eighth"
        with self.assertRaises(ChordNumberOfDotsNotSetError):
            chord.check_printed_duration()
        chord.number_of_dots = 0
        with self.assertRaises(ChordTestError):
            assert chord.check_printed_duration()
        chord.tuplet = Tuplet(3, 2)
        assert chord.check_printed_duration()
        chord.type = "quarter"
        with self.assertRaises(ChordTestError):
            assert chord.check_printed_duration()


class TestDeepcopyChord(ChordTestCase):
    def test_deepcopy_chord_simple(self):
        chord = Chord([60, 62], 2)
        copied = copy.deepcopy(chord)
        assert [midi.value for midi in copied.midis] == [
            midi.value for midi in chord.midis
        ]
        assert [id(midi) for midi in copied.midis] != [id(midi) for midi in chord.midis]
        assert chord.quarter_duration.value == copied.quarter_duration.value
        assert id(chord.quarter_duration) != id(copied.quarter_duration)
        chord._parent = self.mock_beat
        chord.finalize()
        with self.assertRaises(DeepCopyException):
            copy.deepcopy(chord)

    def test_deepcopy_chord_complex(self):
        [
            "_xml_articulations",
            "_xml_direction_types",
            "_xml_directions",
            "_xml_lyrics",
            "_xml_technicals",
            "_xml_ornaments",
            "_xml_dynamics",
            "_xml_other_notations",
            "_note_attributes",
            "_clef",
            "_metronome",
            "_grace_chords",
            "_arpeggio",
            "_after_notes_xml_elements",
            "_beams",
            "_broken_beam",
            "_type",
            "_number_of_dots",
            "_tuplet",
            "_original_starting_ties",
        ]
        chord = Chord([60, 62], 1)
        chord.add_dynamics("p")
        chord.add_lyric("something")
        chord.add_x(XMLAccent())
        chord.add_x(XMLUpBow())
        b = XMLBarline(location="middle")
        b.xml_bar_style = "dashed"
        chord.add_xml_element_after_notes(b)
        chord.add_dynamics("f", placement="above")

        chord.add_grace_chord(midis_or_grace_chord=[60], position="after")
        chord.add_grace_chord(midis_or_grace_chord=[61], position="after")
        chord.add_grace_chord(midis_or_grace_chord=[70], position="before")

        chord.add_tie(type="start")
        chord.add_tie(type="stop")
        chord.add_wedge(wedge="crescendo")
        chord.add_wedge(wedge="stop")
        chord.add_words("something")
        chord.add_words("something else")
        chord.break_beam()

        copied = copy.deepcopy(chord)

        assert [lyric.value_ for lyric in copied._xml_lyrics] == [
            lyric.value_ for lyric in chord._xml_lyrics
        ]
        assert [id(lyric) for lyric in copied._xml_lyrics] != [
            id(lyric) for lyric in chord._xml_lyrics
        ]

        assert isinstance(copied._after_notes_xml_elements[0], XMLBarline)
        assert id(copied._after_notes_xml_elements[0]) != id(
            chord._after_notes_xml_elements[0]
        )

        assert [gch.midis[0].value for gch in copied.get_grace_chords("after")] == [
            60,
            61,
        ]
        assert [gch.midis[0].value for gch in copied.get_grace_chords("before")] == [70]

        assert [id(gch) for gch in copied.get_grace_chords("after")] != [
            id(gch) for gch in chord.get_grace_chords("after")
        ]
        assert [id(gch) for gch in copied.get_grace_chords("before")] != [
            id(gch) for gch in chord.get_grace_chords("before")
        ]

        assert copied.is_tied_to_previous is True
        assert copied.is_tied_to_next is True

        assert [wedge.value_ for wedge in copied.get_wedges()] == [
            wedge.value_ for wedge in chord.get_wedges()
        ]
        assert [id(wedge) for wedge in copied.get_wedges()] != [
            id(wedge) for wedge in chord.get_wedges()
        ]

        assert [word.value_ for word in copied.get_words()] == [
            word.value_ for word in chord.get_words()
        ]
        assert [id(word) for word in copied.get_words()] != [
            id(word) for word in chord.get_words()
        ]

        assert chord.broken_beam == copied.broken_beam



"""Lyrics spanners: continued syllables and extenders, possibly spanning multiple notes. The intermediate notes do
not have any <lyric> element."""

from pathlib import Path

from musicscore import Score, Chord
from musicscore.lyrics import Lyrics
from unittest import TestCase


class TestLily61k(TestCase):
    def test_lily_61h_Lyrics_SprannersExtenders(self):
        score = Score()
        part = score.add_part("p1")

        chords = [Chord(72, 1) for _ in range(11)]
        chords.append(Chord(0, 1))

        Lyrics(
            [("A", None), ("b", None, None, "CC", None, None), ("e", None, None)]
        ).add_to_chords(chords[:-1])

        [part.add_chord(ch) for ch in chords]

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



"""MusicXML allows for overlapping part-groups, while many applications do not allow overlap- ping groups,
but require them to be properly nested. In this case, one group (within parenthesis) goes from staff 1 to 4 and
another group (also within parenthesis) goes from staff 3 to 5."""

from pathlib import Path

from musicscore import Score, Chord
from unittest import TestCase


class TestLily41f(TestCase):
    def test_lily_41f_StaffGroupOverlapping(self):
        score = Score()
        parts = [score.add_part(f"p-{i}") for i in range(1, 6)]
        for p in parts:
            p.add_chord(Chord(0, 4))

        score.group_parts(1, 1, 4, name="Group 1", symbol="bracket")
        score.group_parts(2, 3, 4, name="Group 2", symbol="bracket")

        path = Path(__file__).with_suffix(".xml")
        score.export_xml(path)



import random
from pathlib import Path

from musicscore import Score, QuarterDuration, SimpleFormat, Part
from musicscore.tests.test_metronome import TestCase


class TestQuantization(TestCase):
    def test_complex_quantization(self):
        s = Score()
        random.seed(11)
        quarter_durations = []
        while sum(quarter_durations) < 40:
            quarter_durations.append(
                QuarterDuration(random.random() + random.randint(0, 3))
            )
        quarter_durations.append(QuarterDuration(44 - sum(quarter_durations)))

        p = s.add_child(Part("p1"))
        chords = SimpleFormat(quarter_durations=quarter_durations).chords
        [ch.add_lyric(round(float(ch.quarter_duration), 2)) for ch in chords]
        [ch.add_words(str(i + 1)) for i, ch in enumerate(chords)]
        [p.add_chord(ch) for ch in chords]
        p.get_quantized = True

        p = s.add_child(Part("p2"))
        p.set_possible_subdivisions([8, 4, 2, 3])
        chords = SimpleFormat(quarter_durations=quarter_durations).chords
        [ch.add_words(str(i + 1)) for i, ch in enumerate(chords)]
        [p.add_chord(ch) for ch in chords]
        p.get_quantized = True

        p = s.add_child(Part("p3"))
        p.set_possible_subdivisions([4, 2])
        chords = SimpleFormat(quarter_durations=quarter_durations).chords
        [ch.add_words(str(i + 1)) for i, ch in enumerate(chords)]
        [p.add_chord(ch) for ch in chords]
        p.get_quantized = True

        p = s.add_child(Part("p4"))
        p.set_possible_subdivisions([2])
        chords = SimpleFormat(quarter_durations=quarter_durations).chords
        [ch.add_words(str(i + 1)) for i, ch in enumerate(chords)]
        [p.add_chord(ch) for ch in chords]
        p.get_quantized = True

        p = s.add_child(Part("p5"))
        p.set_possible_subdivisions([1])
        chords = SimpleFormat(quarter_durations=quarter_durations).chords
        [ch.add_words(str(i + 1)) for i, ch in enumerate(chords)]
        [p.add_chord(ch) for ch in chords]
        p.get_quantized = True

        path = Path(__file__).with_suffix(".xml")
        s.export_xml(path=path)



from pathlib import Path

from musicscore import Score, Time, Chord, Key
from unittest import TestCase

"""
Various key signature: from 11 flats to 11 sharps
"""


class TestLily131(TestCase):
    def test_lily_13a_KeySignatures(self):
        score = Score()
        part = score.add_part("p1")
        part.add_measure(Time(2, 4))
        for x in range(-11, 12):
            part.add_chord(Chord(60, 2))
            part.get_current_measure().key = Key(fifths=x)

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from unittest import TestCase, skip

from musicscore import Score
from musicscore.chord import Chord
from musicscore.clef import BassClef, TrebleClef
from musicscore.exceptions import VoiceIsFullError, AddChordError
from musicscore.measure import Measure, generate_measures
from musicscore.part import Part
from musicscore.staff import Staff
from musicscore.time import Time
from musicscore.voice import Voice
from musicxml.xmlelement.xmlelement import *


class TestMeasure(TestCase):
    def test_measure_default_init(self):
        expected = """<measure number="1">
  <attributes>
    <divisions>1</divisions>
    <key>
      <fifths>0</fifths>
    </key>
    <time>
      <beats>4</beats>
      <beat-type>4</beat-type>
    </time>
    <clef>
      <sign>G</sign>
      <line>2</line>
    </clef>
  </attributes>
  <note>
    <rest />
    <duration>4</duration>
    <voice>1</voice>
    <type>whole</type>
  </note>
</measure>
"""
        m = Measure(1)
        m.fill_with_rests()
        assert m.to_string() == expected

    def test_number(self):
        m = Measure(number=2)
        assert m.number == 2
        m.number = 4
        assert m.number == 4
        m.xml_object.number = "3"
        assert m.number == 3

    def test_measure_time_signature(self):
        m = Measure(1)
        expected = """<time>
  <beats>4</beats>
  <beat-type>4</beat-type>
</time>
"""
        assert m.time.to_string() == expected
        m.time = Time(3, 4)
        expected = """<time>
  <beats>3</beats>
  <beat-type>4</beat-type>
</time>
"""
        assert m.time.to_string() == expected
        m = Measure(2, time=Time(5, 8, 2, 4))
        expected = """<time>
  <beats>5</beats>
  <beat-type>8</beat-type>
  <beats>2</beats>
  <beat-type>4</beat-type>
</time>
"""
        assert m.time.to_string() == expected

    def test_measure_add_child_staff(self):
        m = Measure(1)
        assert m.get_children() == []
        st1 = m.add_child(Staff())
        assert st1.number is None
        assert m.get_children() == [st1]
        st2 = m.add_child(Staff())
        assert m.get_children() == [st1, st2]
        assert (st1.number, st2.number) == (1, 2)
        with self.assertRaises(ValueError):
            m.add_child(Staff(number=2))

        m = Measure(1)
        st1 = m.add_child(Staff(number=1))
        assert st1.number == 1

        m = Measure(1)
        with self.assertRaises(ValueError):
            m.add_child(Staff(number=2))

    def test_check_divisions(self):
        m = Measure(1)
        st = m.add_child(Staff())
        v1 = st.add_child(Voice(1))
        v2 = st.add_child(Voice(2))
        v1.update_beats()
        v2.update_beats()
        quarter_durations_1 = [1 / 3, 2 / 3, 2 / 5, 3 / 5, 1, 1 / 2, 1 / 2]
        quarter_durations_2 = [1, 9 / 7, 5 / 7, 2 / 3, 1 / 3]
        for qd in quarter_durations_1:
            v1._add_chord(Chord(60, qd))
        for qd in quarter_durations_2:
            v2._add_chord(Chord(70, qd))
        m._update_divisions()
        assert m.xml_object.xml_attributes.xml_divisions.value_ == 210

    def test_add_chord(self):
        m = Measure(1)
        chord = Chord(60, quarter_duration=2.5)
        chord.midis[0].accidental.show = False
        m._add_chord(chord)
        chord = Chord(61, quarter_duration=1.5)
        chord.midis[0].accidental.show = True
        m._add_chord(chord)
        m.finalize()
        for xml_note, duration in zip(m.find_children("XMLNote"), [4, 1, 1, 2]):
            assert xml_note.xml_duration.value_ == duration
        expected = """<measure number="1">
  <attributes>
    <divisions>2</divisions>
    <key>
      <fifths>0</fifths>
    </key>
    <time>
      <beats>4</beats>
      <beat-type>4</beat-type>
    </time>
    <clef>
      <sign>G</sign>
      <line>2</line>
    </clef>
  </attributes>
  <note>
    <pitch>
      <step>C</step>
      <octave>4</octave>
    </pitch>
    <duration>4</duration>
    <tie type="start" />
    <voice>1</voice>
    <type>half</type>
    <notations>
      <tied type="start" />
    </notations>
  </note>
  <note>
    <pitch>
      <step>C</step>
      <octave>4</octave>
    </pitch>
    <duration>1</duration>
    <tie type="stop" />
    <voice>1</voice>
    <type>eighth</type>
    <beam number="1">begin</beam>
    <notations>
      <tied type="stop" />
    </notations>
  </note>
  <note>
    <pitch>
      <step>C</step>
      <alter>1</alter>
      <octave>4</octave>
    </pitch>
    <duration>1</duration>
    <tie type="start" />
    <voice>1</voice>
    <type>eighth</type>
    <accidental>sharp</accidental>
    <beam number="1">end</beam>
    <notations>
      <tied type="start" />
    </notations>
  </note>
  <note>
    <pitch>
      <step>C</step>
      <alter>1</alter>
      <octave>4</octave>
    </pitch>
    <duration>2</duration>
    <tie type="stop" />
    <voice>1</voice>
    <type>quarter</type>
    <notations>
      <tied type="stop" />
    </notations>
  </note>
</measure>
"""
        assert m.to_string() == expected

    def test_get_staff(self):
        m = Measure(1)
        assert m.get_staff(staff_number=1) is None
        st = m.add_child(Staff())
        assert m.get_staff(staff_number=1) == st
        st.number = 1
        assert m.get_staff(staff_number=1) == st
        st = m.add_child(Staff())
        assert m.get_staff(staff_number=2) == st

    def test_add_staff(self):
        m = Measure(1)
        st1 = m.add_staff()
        assert m.get_children()[0] == st1
        assert m.get_staff(1) == st1
        assert st1.number is None

        st2 = m.add_staff()
        assert st1.number == 1
        assert st2.number == 2
        st3 = m.add_staff()
        assert st3.number == 3
        # print(m.get_children())

        m = Measure(1)
        st = m.add_staff()
        assert st.number is None
        assert m.get_staff(staff_number=1) == st

        st = m.add_staff(1)
        assert m.get_staff(staff_number=1) == st

        st = m.add_staff(5)
        assert m.get_staff(staff_number=5) == st
        assert len(m.get_children()) == 5
        assert [st.number for st in m.get_children()] == [1, 2, 3, 4, 5]

    def test_add_voice(self):
        m = Measure(1)
        v = m.add_voice(staff_number=None, voice_number=2)
        st = m.get_staff(staff_number=1)
        assert st.get_children()[-1] == v
        assert [v.number for v in st.get_children()] == [1, 2]

        m.add_voice(staff_number=2, voice_number=3)
        assert [st.number for st in m.get_children()] == [1, 2]
        assert [v.number for v in m.get_staff(2).get_children()] == [1, 2, 3]

        m = Measure(1)
        m.add_voice(staff_number=2, voice_number=2)
        st1, st2 = m.get_children()
        assert len(st2.get_children()) == 2
        assert len(st1.get_children()) == 1

    def test_get_voice(self):
        m = Measure(1)
        with self.assertRaises(TypeError):
            m.get_voice(None, 1)
        assert m.get_voice(staff_number=1, voice_number=1) is None
        st = m.add_child(Staff())
        assert m.get_voice(staff_number=1, voice_number=1) is None
        v = st.add_child(Voice())
        assert m.get_voice(staff_number=1, voice_number=1) == v
        assert m.get_voice(staff_number=1, voice_number=1) == v
        v = st.add_child(Voice())
        assert m.get_voice(staff_number=1, voice_number=2) == v

    def test_add_chord_voice_filled(self):
        m = Measure(1)
        ch = Chord(quarter_duration=2, midis=60)
        returned_chords = m._add_chord(ch)
        assert len(returned_chords) == 1
        assert returned_chords[0] == ch
        assert returned_chords[0].quarter_duration == 2
        assert m.get_children()[0].get_children()[0].leftover_chord is None
        ch = Chord(quarter_duration=2, midis=60)
        returned_chords = m._add_chord(ch)
        assert len(returned_chords) == 1
        assert returned_chords[0] == ch
        assert returned_chords[0].quarter_duration == 2
        assert m.get_children()[0].get_children()[0].leftover_chord is None
        with self.assertRaises(VoiceIsFullError):
            m._add_chord(Chord(quarter_duration=2, midis=60))
        ch = Chord(quarter_duration=2, midis=60)
        returned_chords = m._add_chord(ch, voice_number=2)
        assert len(returned_chords) == 1
        assert [ch.voice_number for ch in m.get_chords()] == [1, 1, 2]

    def test_add_chord_leftover(self):
        m = Measure(1)
        ch = Chord(quarter_duration=5, midis=60)
        returned_chords = m._add_chord(ch)
        assert len(returned_chords) == 1
        assert returned_chords[0] == ch
        assert returned_chords[0].quarter_duration == 4
        assert (
            m.get_voice(staff_number=1, voice_number=1).leftover_chord.quarter_duration
            == 1
        )

    def test_attributes(self):
        """
        Test that a dot operator can directly reach the xml_object
        """
        m = Measure(1)
        m.xml_object.width = 5
        assert m.width == 5

        m.width = 7
        assert m.xml_object.width == 7

        with self.assertRaises(AttributeError):
            m.hello = "bbb"
        m.xml_object.width = 10

        assert m.xml_object.width == 10
        assert m.width == 10

        m = Measure(1)
        m.width = 10
        assert m.width == 10
        assert m.xml_object.width == 10

    def test_left_and_right_barline_property(self):
        m = Measure(number="1")
        assert (
            m.get_barline(location="left")
            == m.get_barline()
            == m.get_barline(location="right")
            is None
        )

        m.set_barline(location="left")
        assert isinstance(m.get_barline(location="left"), XMLBarline)
        assert not m.get_barline(location="left").get_children()

        m.set_barline(location="right")
        assert isinstance(m.get_barline(location="right"), XMLBarline)
        assert not m.get_barline(location="right").get_children()
        assert isinstance(m.get_barline(), XMLBarline)
        assert not m.get_barline().get_children()

        barline_styles = [
            "regular",
            "regular",
            "dotted",
            "dashed",
            "heavy",
            "light-light",
            "light-heavy",
            "heavy-light",
            "heavy-heavy",
            "tick",
            "short",
            "none",
        ]
        for bs in barline_styles:
            m = Measure(1)
            m.set_barline(style=bs)
            m.finalize()
            assert m.xml_barline.xml_bar_style.value_ == bs
            assert m.xml_barline.location == "right"

        for bs in barline_styles:
            m = Measure(1)
            m.set_barline(style=bs, location="left")
            m.finalize()
            assert m.xml_barline.xml_bar_style.value_ == bs
            assert m.xml_barline.location == "left"

    def test_barline(self):
        m = Measure(number="1")
        m.set_barline(style="light-light")
        expected = """<measure number="1">
  <attributes>
    <divisions>1</divisions>
    <key>
      <fifths>0</fifths>
    </key>
    <time>
      <beats>4</beats>
      <beat-type>4</beat-type>
    </time>
    <clef>
      <sign>G</sign>
      <line>2</line>
    </clef>
  </attributes>
  <note>
    <rest />
    <duration>4</duration>
    <voice>1</voice>
    <type>whole</type>
  </note>
  <barline location="right">
    <bar-style>light-light</bar-style>
  </barline>
</measure>
"""
        m.fill_with_rests()
        assert m.to_string() == expected

    def test_update_beats_from_parent_measure(self):
        v = Voice()
        m = Measure(1)
        st = m.add_child(Staff())
        st.add_child(v)
        v.update_beats()
        assert [
            child.quarter_duration.as_integer_ratio() for child in v.get_children()
        ] == [(1, 1)] * 4

        m = Measure(1, time=Time(3, 4, 1, 8))
        st = m.add_child(Staff())
        st.add_child(v)
        v.update_beats()
        assert [
            child.quarter_duration.as_integer_ratio() for child in v.get_children()
        ] == [(1, 1)] * 3 + [(1, 2)]

        m.time = Time(2, 8)
        assert [
            child.quarter_duration.as_integer_ratio() for child in v.get_children()
        ] == [(1, 1)]

        m.time.signatures = [3, 4, 1, 8]
        assert [
            child.quarter_duration.as_integer_ratio() for child in v.get_children()
        ] == [(1, 1)] * 3 + [(1, 2)]

        m.time.actual_signatures = [1, 8, 1, 8]
        assert [
            child.quarter_duration.as_integer_ratio() for child in v.get_children()
        ] == [(1, 2)] * 2

    def test_generate_measures(self):
        times = [2 * Time(3, 8), (3, 4), 3 * [(1, 8)], Time(1, 8, 3, 4), Time(3, 4)]
        measures = generate_measures(times, 3)
        assert len(measures) == 8
        assert [m.time.signatures for m in measures] == [
            (3, 8),
            (3, 8),
            (3, 4),
            (1, 8),
            (1, 8),
            (1, 8),
            (1, 8, 3, 4),
            (3, 4),
        ]
        assert [m.number for m in measures] == list(range(3, 11))

    def test_add_chord_staff_and_voice(self):
        m = Measure(1)
        m._add_chord(Chord(60, 1), staff_number=2, voice_number=2)
        assert m.get_staff(2) is not None
        assert m.get_voice(staff_number=2, voice_number=2) is not None
        m._add_chord(Chord(60, 1), staff_number=4, voice_number=5)
        assert m.get_voice(staff_number=3, voice_number=1) is not None
        assert m.get_voice(staff_number=3, voice_number=2) is None
        assert m.get_voice(staff_number=3, voice_number=5) is None
        assert m.get_voice(staff_number=4, voice_number=1) is not None
        assert m.get_voice(staff_number=4, voice_number=2) is not None
        assert m.get_voice(staff_number=4, voice_number=3) is not None
        assert m.get_voice(staff_number=4, voice_number=4) is not None
        assert m.get_voice(staff_number=4, voice_number=5) is not None

    def test_update_dynamics(self):
        m = Measure(1)
        chord = Chord(60, quarter_duration=2.5)
        chord.add_dynamics(["ppp", "fff"])
        m._add_chord(chord)
        m.fill_with_rests()
        m.finalize()
        assert m.xml_direction is not None
        expected = """<direction placement="below">
    <direction-type>
      <dynamics>
        <ppp />
      </dynamics>
    </direction-type>
    <direction-type>
      <dynamics>
        <fff />
      </dynamics>
    </direction-type>
  </direction>
"""

        assert m.xml_direction.to_string() == expected

    def test_add_chord_to_measure(self):
        part = Part(id="part-1")
        measure = part.add_measure(time=Time(4, 4))
        with self.assertRaises(AddChordError):
            measure.add_chord(Chord(midis=60, quarter_duration=15))

    def test_default_staff(self):
        # inconsistency add_part doesn't add measure but add_measure does create staff, voice and beats
        m = Measure(1)
        assert len(m.get_children()) == 0
        score = Score()
        p = score.add_part("p1")
        assert len(m.get_children()) == 0
        m = p.add_measure()
        assert len(m.get_children()) == 1
        assert len(m.get_children()[0].get_children()) == 1
        v = m.get_children()[0].get_children()[0]
        assert len(v.get_children()) == 4

    def test_new_system(self):
        s = Score()
        p = s.add_part("part1")
        p.add_chord(Chord(quarter_duration=4 * 4 * 4, midis=60))
        m = p.get_measure(1)
        assert m.new_system is False
        m.new_system = True
        assert m.new_system is True
        assert m.xml_object.xml_print.new_system == "yes"
        s.finalize()
        s.xml_object._final_checks()

    def test_add_repeat_left(self):
        m = Measure(1)
        m._add_chord(Chord(60, 4))
        m.set_repeat_barline(times=5)
        expected = """<measure number="1">
  <attributes>
    <divisions>1</divisions>
    <key>
      <fifths>0</fifths>
    </key>
    <time>
      <beats>4</beats>
      <beat-type>4</beat-type>
    </time>
    <clef>
      <sign>G</sign>
      <line>2</line>
    </clef>
  </attributes>
  <note>
    <pitch>
      <step>C</step>
      <octave>4</octave>
    </pitch>
    <duration>4</duration>
    <voice>1</voice>
    <type>whole</type>
  </note>
  <barline location="right">
    <bar-style>light-heavy</bar-style>
    <repeat direction="backward" times="5" />
  </barline>
</measure>
"""
        assert m.to_string() == expected

    def test_add_ending_without_repeat(self):
        m = Measure(1)
        # add_ending needs number and type
        with self.assertRaises(TypeError):
            m.set_repeat_ending()
        with self.assertRaises(TypeError):
            m.set_repeat_ending(number=1)
        with self.assertRaises(TypeError):
            m.set_repeat_ending(type="stop")
        # wrong type
        with self.assertRaises(ValueError):
            m.set_repeat_ending(number=1, type="bal")

        # right init
        m._add_chord(Chord(0, 4))
        m.set_repeat_ending(number=2, type="discontinue")
        expected = """<measure number="1">
  <attributes>
    <divisions>1</divisions>
    <key>
      <fifths>0</fifths>
    </key>
    <time>
      <beats>4</beats>
      <beat-type>4</beat-type>
    </time>
    <clef>
      <sign>G</sign>
      <line>2</line>
    </clef>
  </attributes>
  <note>
    <rest />
    <duration>4</duration>
    <voice>1</voice>
    <type>whole</type>
  </note>
  <barline location="right">
    <ending number="2" type="discontinue" />
  </barline>
</measure>
"""
        assert m.to_string() == expected

    def test_measure_fill_with_rests(self):
        m = Measure(1)
        m.fill_with_rests()
        ch = m.get_chords()[0]
        assert ch.is_rest
        assert ch.quarter_duration == 4


class TestUpdateAccidentals(TestCase):
    def test_update_accidentals_simple(self):
        m = Measure(1)
        midis = [60, 61, 62, 60]
        for mi in midis:
            m._add_chord(Chord(mi, quarter_duration=1))
        m.finalize()
        assert [ch.midis[0].accidental.show for ch in m.get_chords()] == [
            False,
            True,
            False,
            True,
        ]

    def test_update_accidentals_with_last_steps(self):
        p = Part("P1")
        p.add_measure()
        p.add_measure()
        p.add_chord(Chord(midis=60, quarter_duration=2))
        p.add_chord(Chord(midis=61, quarter_duration=2))
        p.add_chord(Chord(midis=60, quarter_duration=4))
        for m in p.get_children():
            m.finalize()
        last_chord = p.get_children()[-1].get_children()[-1].get_chords()[0]
        assert last_chord.midis[0].accidental.sign == "natural"
        assert last_chord.midis[0].accidental.show

    def test_quarter_duration(self):
        m = Measure(1)
        assert m.quarter_duration == 4
        m.time.signatures = (3, 4)
        assert m.quarter_duration == 3
        m.time.signatures = (2, 4, 1, 8)
        assert m.quarter_duration == 2.5

    def test_xml_notes_with_different_voices(self):
        m = Measure(1)
        m._add_chord(Chord(60, 4))
        m._add_chord(Chord(60, 4), voice_number=2)
        m.finalize()
        ch1, ch2 = m.get_chords()
        assert ch1.notes[0].xml_voice.value_ == "1"
        assert ch2.notes[0].xml_voice.value_ == "2"
        b = m.xml_object.find_child("XMLBackup")
        assert b is not None

    def test_xml_notes_with_different_staves(self):
        m = Measure(1)
        m._add_chord(Chord(72, 4), staff_number=1, voice_number=1)
        m._add_chord(Chord(60, 4), staff_number=1, voice_number=2)
        m._add_chord(Chord(48, 4), staff_number=2, voice_number=1)
        m._add_chord(Chord(36, 4), staff_number=2, voice_number=2)
        m.clefs[1] = BassClef()
        m.finalize()
        ch1, ch2, ch3, ch4 = m.get_chords()
        assert ch1.notes[0].xml_staff.value_ == 1
        assert ch2.notes[0].xml_staff.value_ == 1
        assert ch3.notes[0].xml_staff.value_ == 2
        assert ch4.notes[0].xml_staff.value_ == 2
        assert ch1.notes[0].xml_voice.value_ == "1"
        assert ch2.notes[0].xml_voice.value_ == "2"
        assert ch3.notes[0].xml_voice.value_ == "1"
        assert ch4.notes[0].xml_voice.value_ == "2"
        backups = m.xml_object.find_children("XMLBackup")
        assert len(backups) == 3
        assert m.xml_attributes.xml_staves.value_ == 2

        cl1, cl2 = m.xml_attributes.find_children("XMLClef")
        assert cl1.number == 1
        assert cl2.number == 2
        assert cl1.xml_sign.value_ == "G"
        assert cl1.xml_line.value_ == 2
        assert cl2.xml_sign.value_ == "F"
        assert cl2.xml_line.value_ == 4

    @skip
    def test_update_chord_accidentals(self):
        self.fail()


class TestMeasureAttributes(TestCase):
    def test_measure_update_attributes(self):
        expected = """<measure number="1">
  <attributes>
    <divisions>1</divisions>
    <key>
      <fifths>0</fifths>
    </key>
    <time>
      <beats>4</beats>
      <beat-type>4</beat-type>
    </time>
    <clef>
      <sign>G</sign>
      <line>2</line>
    </clef>
  </attributes>
  <note>
    <rest />
    <duration>4</duration>
    <voice>1</voice>
    <type>whole</type>
  </note>
</measure>
"""
        m = Measure(1)
        m.add_staff()
        m._update_attributes()
        m.fill_with_rests()
        assert m.to_string() == expected

    def test_measure_clefs(self):
        m = Measure(1)
        m.add_staff()
        m.finalize()
        clefs = m.xml_object.xml_attributes.find_children("XMLClef")
        assert clefs[0].xml_sign.value_ == "G"
        assert clefs[0].xml_line.value_ == 2

        m = Measure(1)
        m.add_staff()
        m.add_staff()
        m.finalize()
        clefs = m.xml_object.xml_attributes.find_children("XMLClef")
        assert clefs[0].xml_sign.value_ == "G"
        assert clefs[0].xml_line.value_ == 2
        assert clefs[1].xml_sign.value_ == "F"
        assert clefs[1].xml_line.value_ == 4

        m = Measure(1)
        m.add_staff()
        m.add_staff()
        m.finalize()
        m.clefs[0].sign = "C"
        m.clefs[0].line = 3

        clefs = m.xml_object.xml_attributes.find_children("XMLClef")
        assert clefs[0].xml_sign.value_ == "C"
        assert clefs[0].xml_line.value_ == 3

    def test_measure_add_staff_clef(self):
        m = Measure(1)
        assert not m.clefs
        m.add_staff()
        assert len(m.clefs) == 1
        assert m.clefs[0].number is None
        m.clefs[0].sign = "G"
        m.clefs[0].line = 2

        m.add_staff()
        assert len(m.clefs) == 2
        assert [cl.number for cl in m.clefs] == [1, 2]
        assert [cl.sign for cl in m.clefs] == ["G", "F"]
        assert [cl.line for cl in m.clefs] == [2, 4]

        m.add_staff()
        assert len(m.clefs) == 3
        assert [cl.number for cl in m.clefs] == [1, 2, 3]
        assert [cl.sign for cl in m.clefs] == ["G", "G", "F"]
        assert [cl.line for cl in m.clefs] == [2, 2, 4]
        assert [cl.octave_change for cl in m.clefs] == [2, None, None]

        m.add_staff()
        assert len(m.clefs) == 4
        assert [cl.number for cl in m.clefs] == [1, 2, 3, 4]
        assert [cl.sign for cl in m.clefs] == ["G", "G", "F", "F"]
        assert [cl.line for cl in m.clefs] == [2, 2, 4, 4]
        assert [cl.octave_change for cl in m.clefs] == [2, None, None, -2]

        m.add_staff()
        assert len(m.clefs) == 5
        assert [cl.number for cl in m.clefs] == [1, 2, 3, 4, 5]
        assert [cl.sign for cl in m.clefs] == ["G", "G", "G", "F", "F"]
        assert [cl.line for cl in m.clefs] == [2, 2, 2, 4, 4]
        assert [cl.octave_change for cl in m.clefs] == [2, None, None, None, -2]

        m.add_staff()
        assert len(m.clefs) == 6
        assert [cl.number for cl in m.clefs] == [1, 2, 3, 4, 5, 6]
        assert [cl.sign for cl in m.clefs] == ["G", "G", "G", "F", "F", "F"]
        assert [cl.line for cl in m.clefs] == [2, 2, 2, 4, 4, 4]
        assert [cl.octave_change for cl in m.clefs] == [2, None, None, None, None, -2]

    def test_measure_add_staff_and_change_clef_manually(self):
        m = Measure(1)
        m.add_staff()
        assert m.get_staff(1).clef.number is None
        bass_clef = BassClef()
        m.get_children()[0].clef = bass_clef
        assert m.get_staff(1).clef.number is None
        assert m.get_staff(1).clef == bass_clef

        m.add_staff()
        assert m.get_staff(1).clef.number == 1
        assert m.get_staff(2).clef.number == 2
        m.get_staff(2).clef = TrebleClef()
        assert m.get_staff(2).clef.number == 2

        assert m.get_staff(1).clef == bass_clef

        treble_clef = TrebleClef()
        m.add_child(Staff(number=3, clef=treble_clef))
        assert m.get_staff(3).clef.number == 3
        assert m.get_staff(3).clef == treble_clef

    def test_measure_false_show_keys(self):
        expected = """<measure number="1">
  <attributes>
    <divisions>1</divisions>
    <time>
      <beats>4</beats>
      <beat-type>4</beat-type>
    </time>
    <clef>
      <sign>G</sign>
      <line>2</line>
    </clef>
  </attributes>
  <note>
    <rest />
    <duration>4</duration>
    <voice>1</voice>
    <type>whole</type>
  </note>
</measure>
"""
        m = Measure(1)
        m.add_staff()
        m.key.show = False
        m._update_attributes()
        m.fill_with_rests()
        assert m.to_string() == expected

    def test_measure_false_show_time(self):
        expected = """<measure number="1">
  <attributes>
    <divisions>1</divisions>
    <key>
      <fifths>0</fifths>
    </key>
    <clef>
      <sign>G</sign>
      <line>2</line>
    </clef>
  </attributes>
  <note>
    <rest />
    <duration>4</duration>
    <voice>1</voice>
    <type>whole</type>
  </note>
</measure>
"""
        m = Measure(1)
        m.add_staff()
        m.time.show = False
        m._update_attributes()
        m.fill_with_rests()
        assert m.to_string() == expected

    def test_measure_false_show_clef(self):
        expected = """<measure number="1">
  <attributes>
    <divisions>1</divisions>
    <key>
      <fifths>0</fifths>
    </key>
    <time>
      <beats>4</beats>
      <beat-type>4</beat-type>
    </time>
  </attributes>
  <note>
    <rest />
    <duration>4</duration>
    <voice>1</voice>
    <type>whole</type>
  </note>
</measure>
"""
        m = Measure(1)
        m.add_staff()
        m.clefs[0].show = False
        m.fill_with_rests()
        assert m.to_string() == expected


class TestMeasureBeatGrouping(TestCase):
    def test_measure_5_4(self):
        m = Measure(1, time=Time(5, 4))
        chords = m._add_chord(Chord(midis=60, quarter_duration=5))
        assert [ch.quarter_duration for ch in chords] == [3, 2]

    def test_measure_6_4(self):
        m = Measure(1, time=Time(6, 4))
        chords = m._add_chord(Chord(midis=60, quarter_duration=6))
        assert [ch.quarter_duration for ch in chords] == [6]

    def test_add_staff_to_score(self):
        score = Score()
        part = score.add_part("p1")
        measure = part.add_measure()
        st = measure.add_staff()
        assert st == measure.get_children()[-1]



from typing import Optional

from musicscore.xmlwrapper import XMLWrapper
from musicxml.xmlelement.xmlelement import XMLClef

__all__ = ["Clef", "TrebleClef", "BassClef", "AltoClef", "TenorClef"]


class Clef(XMLWrapper):
    _ATTRIBUTES = {"show", "sign", "line", "octave_change"}
    XMLClass = XMLClef

    def __init__(
        self,
        sign: str = "G",
        line: Optional[int] = 2,
        octave_change: int = None,
        show: bool = True,
        default: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__()
        self._xml_object = self.XMLClass(*args, **kwargs)
        self._show = None
        self.show = show
        self.line = line
        self.sign = sign
        self.octave_change = octave_change

        self._default = default

    @property
    def line(self) -> Optional[int]:
        """
        Set and get ``value_`` of :obj:`~musicxml.xmlelement.xmlelement.XMLLine`

        :return: ``self.xml_object.xml_line.value_``
        :rtype: int, None
        """
        if self.xml_object.xml_line:
            return self.xml_object.xml_line.value_

    @line.setter
    def line(self, val):
        self.xml_object.xml_line = val

    @property
    def octave_change(self) -> Optional[int]:
        """
        Set and get ``value_`` (int) of :obj:`~musicxml.xmlelement.xmlelement.XMLOctaveChange` child object of associated :obj:`~musicxml.xmlelement.xmlelement.XMLClef` which indicates how many octaves must be added to get from written pitch to the sounding pitch.

        :return: ``self.xml_object.xml_clef_octave_change.value_``
        """

        if self.xml_object.xml_clef_octave_change:
            return self.xml_object.xml_clef_octave_change.value_

    @octave_change.setter
    def octave_change(self, val):
        self.xml_object.xml_clef_octave_change = val

    @property
    def sign(self) -> Optional[str]:
        """
        Set and get ``value_`` of :obj:`~musicxml.xmlelement.xmlelement.XMLSign` child object of associated :obj:`~musicxml.xmlelement.xmlelement.XMLClef`

        :return: ``self.xml_object.xml_sign.value_``
        :rtype: str, None
        """
        if self.xml_object.xml_sign:
            return self.xml_object.xml_sign.value_

    @sign.setter
    def sign(self, val):
        self.xml_object.xml_sign = val

    @property
    def show(self) -> bool:
        """
        Set and get show attribute. If ``False`` :obj:`Clef` element is not shown.
        """
        return self._show

    @show.setter
    def show(self, val):
        if not isinstance(val, bool):
            raise TypeError
        self._show = val

    def __copy__(self):
        new_clef = self.__class__()
        new_clef.sign = self.sign
        new_clef.line = self.line
        new_clef.octave_change = self.octave_change
        new_clef.show = self._show
        return new_clef


class TrebleClef(Clef):
    """
    Default parameters:
      - sign='G'
      - line=2
    """

    def __init__(
        self, show: bool = True, octave_change: Optional[int] = None, **kwargs
    ):
        super().__init__(
            sign="G", line=2, show=show, octave_change=octave_change, **kwargs
        )


class BassClef(Clef):
    """
    Default parameters:
      - sign='F'
      - line=4
    """

    def __init__(
        self, show: bool = True, octave_change: Optional[int] = None, **kwargs
    ):
        super().__init__(
            sign="F", line=4, show=show, octave_change=octave_change, **kwargs
        )


class AltoClef(Clef):
    """
    Default parameters:
      - sign='C'
      - line=3
    """

    def __init__(
        self, show: bool = True, octave_change: Optional[int] = None, **kwargs
    ):
        super().__init__(
            sign="C", line=3, show=show, octave_change=octave_change, **kwargs
        )


class TenorClef(Clef):
    """
    Default parameters:
      - sign='C'
      - line=4
    """

    def __init__(
        self, show: bool = True, octave_change: Optional[int] = None, **kwargs
    ):
        super().__init__(
            sign="C", line=4, show=show, octave_change=octave_change, **kwargs
        )


class PercussionClef(Clef):
    """
    Default parameters:
      - sign=percussion
      - line=None
    """

    def __init__(
        self, show: bool = True, octave_change: Optional[int] = None, **kwargs
    ):
        super().__init__(
            sign="percussion",
            line=None,
            show=show,
            octave_change=octave_change,
            **kwargs,
        )



"""
Different types of (non-repeat) barlines: default (no setting), regular, dotted, dashed, heavy, light-light, light-heavy,
heavy-light, heavy-heavy, tick, short, none.
"""

from pathlib import Path

from musicscore import Score, Chord
from unittest import TestCase
from musicxml.xmlelement.xmlelement import XMLWords


class TestLily46a(TestCase):
    def test_lily_46a_Barlines(self):
        score = Score()
        p = score.add_part("p1")
        p.add_chord(Chord(0, 4))

        barline_styles = [
            "regular",
            "dotted",
            "dashed",
            "heavy",
            "light-light",
            "light-heavy",
            "heavy-light",
            "heavy-heavy",
            "tick",
            "short",
            "none",
        ]

        for bs in barline_styles:
            ch = Chord(0, 4)
            ch.add_x(XMLWords(bs), placement="below")
            p.add_chord(ch)
            p.get_current_measure().set_barline(style=bs)

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



import inspect
import xml.etree.ElementTree as ET
from difflib import Differ
from pathlib import Path
from unittest import TestCase
from unittest.mock import Mock, patch

import xmltodict
from deepdiff import DeepDiff

from musicscore import generate_measures, Chord, C
from musicxml.xmlelement.xmlelement import *  # noqa: F403
from musicscore.util import (
    XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS,
    XML_DIRECTION_TYPE_CLASSES,
    XML_OTHER_NOTATIONS,
    XML_ORNAMENT_CLASSES,
    XML_ARTICULATION_CLASSES,
    XML_TECHNICAL_CLASSES,
    XML_DYNAMIC_CLASSES,
    XML_ORNAMENT_AND_OTHER_NOTATIONS,
)


class XMLsDifferException(Exception):
    pass


notehead_values = [
    "slash",
    "triangle",
    "diamond",
    "square",
    "rectangle",
    "cross",
    "x",
    "circle dot",
    "circle-x",
    "circled",
    "inverted triangle",
    "left triangle",
    "arrow down",
    "arrow up",
    "slashed",
    "back slashed",
    "normal",
    "cluster",
    "none",
]
notehead_aikin_values = ["do", "re", "mi", "fa", "fa up", "so", "la", "ti"]


def check_notes(notes, midi_values, quarter_durations):
    if len(notes) != len(midi_values) or len(notes) != len(quarter_durations):
        raise ValueError
    for n, m, qd in zip(notes, midi_values, quarter_durations):
        assert n.midi.value == m
        assert n.quarter_duration == qd


class ChordTestCase(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.mock_beat = Mock()
        self.mock_voice = Mock()
        self.mock_staff = Mock()
        self.mock_measure = Mock()

        self.mock_voice.number = 1
        self.mock_beat.up = self.mock_voice
        self.mock_voice.up = self.mock_staff
        self.mock_staff.up = self.mock_measure
        self.mock_measure.get_divisions.return_value = 1
        self.mock_staff.number = None

    def tearDown(self) -> None:
        patch.stopall()


def _create_expected_path(path):
    return Path(path.parent / f"{path.stem}_expected{path.suffix}")


def diff_xml(path_1, path_2=None):
    if path_2 is None:
        path_2 = _create_expected_path(path_1)

    with open(path_1) as f1, open(path_2) as f2:
        f_1 = [i.strip() for i in f1]
        f_2 = [i.strip() for i in f2]

    diff = Differ()
    difference = list(diff.compare(f_1, f_2))
    return [d for d in difference if d.startswith("-") or d.startswith("+")]


def get_xml_elements_diff(el1, el2):
    return DeepDiff(
        xmltodict.parse(ET.tostring(el1)), xmltodict.parse(ET.tostring(el2))
    )


def get_xml_diff_part(expected, xml_path, file_path):
    el1 = ET.parse(file_path.parent / xml_path).getroot().find("part[@id='part-1']")
    el2 = ET.parse(file_path.parent / expected).getroot().find("part[@id='part-1']")
    diff = get_xml_elements_diff(el1=el1, el2=el2)
    if diff:
        raise XMLsDifferException(diff)


def generate_xml_file(score, *simpleformats, path):
    part = score.add_part(id="part-1")
    for index, simpleformat in enumerate(simpleformats):
        for chord in simpleformat.chords:
            part.add_chord(chord, staff_number=index + 1)
    score.export_xml(path)


def generate_repetitions(part):
    for m in generate_measures(
        [(3, 8), (3, 4)] + 2 * [(6, 4)] + 3 * [(4, 4)] + [(5, 4)]
    ):
        part.add_child(m)

    for qd in (
        3 * [0.5]
        + 3 * [1]
        + 4 * [1.5]
        + 3 * [2]
        + [1, 1, 1, 2, 1, 1, 2, 2, 3, 3, 3, 4, 4, 2, 2, 1.5, 1.5, 3.25, 1.75, 1]
    ):
        part.add_chord(Chord(C(5, "#"), qd))


def generate_path(frame):
    f = str(inspect.getframeinfo(frame).function) + ".xml"
    path = Path(inspect.getframeinfo(frame).filename).parent / f
    return path


def create_test_xml_paths(path, test_name):
    return path.parent.joinpath(f"{path.stem}_{test_name}.xml"), path.parent.joinpath(
        f"{path.stem}_{test_name}_expected.xml"
    )


def create_test_objects(type):
    output = []
    if type == "direction_type":
        for cl in XML_DIRECTION_TYPE_CLASSES + XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS:
            if cl == XMLSymbol:
                obj = cl("0")
            elif cl == XMLWedge:
                obj = cl(type="crescendo")
            elif cl == XMLDashes:
                obj = cl(type="start")
            elif cl == XMLBracket:
                obj = cl(type="start", line_end="none")
            elif cl == XMLPedal:
                obj = cl(type="start")
            elif cl == XMLMetronome:
                obj = cl()
                obj.add_child(XMLBeatUnit("quarter"))
                obj.add_child(XMLPerMinute("120"))
            elif cl == XMLOctaveShift:
                obj = cl(type="up")
            elif cl == XMLHarpPedals:
                obj = cl()
                pt = obj.add_child(XMLPedalTuning())
                pt.add_child(XMLPedalStep("A"))
                pt.add_child(XMLPedalAlter(1))
            elif cl == XMLStringMute:
                obj = cl(type="on")
            elif cl == XMLScordatura:
                obj = cl()
                acc = obj.add_child(XMLAccord())
                acc.add_child(XMLTuningStep("A"))
                acc.add_child(XMLTuningOctave(0))
            # elif cl == XMLImage:
            #     obj = cl(source='www.example.com', type='image/gif')
            elif cl == XMLPrincipalVoice:
                obj = cl(type="start", symbol="none")
            elif cl == XMLPercussion:
                obj = cl()
                obj.add_child(XMLWood("cabasa"))
            elif cl == XMLStaffDivide:
                obj = cl(type="up")
            else:
                obj = cl()
            output.append(obj)

    elif type == "technical":
        needed_values = {
            XMLFret: 1,
            XMLString: 1,
            XMLHandbell: "belltree",
            XMLFingering: "2",
            XMLPluck: "something",
            XMLTap: "2",
            XMLHarmonClosed: "yes",
            XMLOtherTechnical: "something",
        }
        needed_types = {XMLHammerOn: "start", XMLPullOff: "start"}
        needed_children = {
            XMLBend: XMLBendAlter(2),
            XMLHole: XMLHoleClosed("yes"),
            XMLArrow: XMLArrowDirection("up"),
            XMLHarmonMute: XMLHarmonClosed("yes"),
        }
        for cl in XML_TECHNICAL_CLASSES:
            if cl in needed_children:
                obj = cl()
                obj.add_child(needed_children[cl])
            elif cl in needed_types:
                obj = cl(type=needed_types[cl])
            elif cl in needed_values:
                obj = cl(needed_values[cl])
            else:
                obj = cl()
            output.append(obj)

    elif type == "ornament":
        for cl in XML_ORNAMENT_AND_OTHER_NOTATIONS + XML_ORNAMENT_CLASSES:
            if cl == XMLAccidentalMark:
                obj = cl("sharp")
            elif cl == XMLTremolo:
                obj = cl(3)
            else:
                obj = cl()
            output.append(obj)

    elif type == "notation":
        for cl in XML_OTHER_NOTATIONS + XML_ORNAMENT_AND_OTHER_NOTATIONS:
            if cl == XMLAccidentalMark:
                obj = cl("sharp")
            else:
                obj = cl()
            output.append(obj)

    elif type == "articulation":
        for cl in XML_ARTICULATION_CLASSES:
            obj = cl()
            output.append(obj)
    elif type == "dynamics":
        for cl in XML_DYNAMIC_CLASSES:
            obj = cl()
            output.append(obj)
    else:
        raise NotImplementedError(f"type {type}")
    return output


def create_test_path(path, test_name):
    return path.parent.joinpath(f"{path.stem}_{test_name}")


class FilePath:
    def __init__(self, unittest, parent_path, name, extension):
        self._unittest = None
        self._parent_path = None
        self.unittest = unittest
        self.parent_path = parent_path
        self.name = name
        self.extension = extension
        self.out_path = None

    @property
    def unittest(self):
        return self._unittest

    @unittest.setter
    def unittest(self, val):
        self._unittest = val

    @property
    def parent_path(self):
        return self._parent_path

    @parent_path.setter
    def parent_path(self, val):
        if not isinstance(val, Path):
            raise TypeError(
                f"parent_path.value must be of type {type(Path)} not{type(val)}"
            )
        self._parent_path = val

    def __enter__(self):
        self.out_path = create_test_path(
            self.parent_path, self.name + "." + self.extension
        )
        return self.out_path

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unittest.assertCompareFiles(self.out_path)


class XMLTestCase(TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _compare_contents(self, actual_file_path, expected_file_path):
        with open(actual_file_path, "r") as myfile:
            result = myfile.read()

        with open(expected_file_path, "r") as myfile:
            expected = myfile.read()

        self.assertEqual(expected, result)

    def assertCompareFiles(self, actual_file_path, expected_file_path=None):
        if not expected_file_path:
            expected_file_path = (
                actual_file_path.parent
                / f"{actual_file_path.stem}_expected{actual_file_path.suffix}"
            )

        self._compare_contents(actual_file_path, expected_file_path)

    def file_path(self, parent_path, name):
        tfp = FilePath(self, parent_path, name, "xml")
        return tfp


def _generate_xml_lyric(text=None, number=1, syllabic=None, extend=None, **kwargs):
    xl = XMLLyric(number=str(number), **kwargs)
    xl.xml_text = text
    xl.xml_syllabic = syllabic
    if extend:
        xl.xml_extend = XMLExtend(type=extend)
    return xl



from pathlib import Path

from musicscore import Score, Chord
from unittest import TestCase


class TestParts(TestCase):
    def test_multiple_parts(self):
        """
        Create multiple parts with names
        """
        score = Score()
        for i in range(5):
            p = score.add_part(f"part-{i}")
            p.add_chord(Chord(0, 4))

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)



from pathlib import Path

from musicscore import Score, Chord, Midi, Time
from unittest import TestCase

"""
All pitch intervals in ascending jump size.
"""


class TestLily01b(TestCase):
    def test_lily_01b_Pitches_Intervals(self):
        score = Score("Various pitches and interval sizes")
        p = score.add_part("intervals")
        p.add_measure(Time(2, 4))
        for i in range(24):
            m1 = Midi(72 + i)
            m2 = Midi(72 - i)
            p.add_chord(Chord(m1, 1))
            p.add_chord(Chord(m2, 1))

        xml_path = Path(__file__).with_suffix(".xml")
        score.export_xml(xml_path)
